"""
Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Module defining io operations performed on restart files.

Public Classes:
- N/A

Functions:

- `clone_restart_file`, helper function to clone an existing restart file.
    Mostly useful for testing purposes of our restart-file writing
    (otherwise we would use copy).
- `open_restart_file`, open a restart file and return a RestartFile object.
- `reassemble_restart_file_from_parts`, reassemble a complete restart file from
    parts. Part of a reduce operation after we conducted operations on part
    of the restart file in parallel (different instances).
- `write_partial_restart_file`, write a partial restart file as output. Part of
    a map operation, where we only operate on a small part of the restart file.
- `write_restart_file`, write a new restart file given updated spectra.

How To Use This Module
======================
(See the individual functions for details.)

1. import
2. open a restart file with open_restart_file
3. perform any of the read write operations using write_restart_file,
   write_partial_restart_file or reassemble_restart_file_from_parts
"""


import numpy
from roguewave.wavewatch3.model_definition import read_model_definition
from roguewave.wavewatch3.resources import create_resource
from roguewave.wavewatch3.restart_file import RestartFile, RestartFileStack
from roguewave.wavewatch3.restart_file_metadata import read_header
from multiprocessing.pool import ThreadPool
from tqdm import tqdm
from typing import Union
from xarray import Dataset,DataArray


def open_restart_file( restart_file, model_definition_file) -> RestartFile:
    """
    Open a restart file locally or remote and return a restart file object.

    :param restart_file: path or uri to a valid restart file
    :param model_definition_file: path or uri to model definition file that
        corresponds to the restart file.
    :return: A restart file object
    """
    restart_file_resource = create_resource(restart_file)
    model_definition_resource = create_resource(model_definition_file)

    meta_data = read_header(restart_file_resource)
    grid, depth, mask = read_model_definition(model_definition_resource)
    return RestartFile(grid, meta_data, restart_file_resource, depth)

def open_restart_file_stack( uris, model_definition_file,
                             cache=False,cache_name=None) -> RestartFileStack:
    """
    Open a restart file locally or remote and return a restart file object.

    :param restart_file: path or uri to a valid restart file
    :param model_definition_file: path or uri to model definition file that
        corresponds to the restart file.
    :return: A restart file object
    """
    restart_files = []
    model_definition_resource = create_resource(model_definition_file,'rb',
                                                cache,cache_name)
    grid, depth, mask = read_model_definition(model_definition_resource)

    for uri in uris:
        restart_file_resource = create_resource(uri,'rb',cache,cache_name)
        meta_data = read_header(restart_file_resource)
        restart_files.append(
            RestartFile(grid, meta_data, restart_file_resource, depth)
        )

    return RestartFileStack(restart_files)


def reassemble_restart_file_from_parts( target_file,
                                        locations,
                                        source_restart_file):
    """
    Reassemble a valid restart file from partial restart files and write the
    result to the target file. This is used as part of a reduce operation to
    create a valid restart file from partial files created in batches in a map
    operation.

    :param target_file: Location to construct new restart file.
    :param locations:  Locations (remote or local) to read partial restart
        files from.
    :param model_definition_file: Location (remote or local) of model
        definition file.
    :return: None
    """

    # Open the partial restart files
    data = []

    def _worker(location):
        partial_spectra_reader = _PartialRestartFileReader(location,
                                                           source_restart_file)
        return partial_spectra_reader.start, partial_spectra_reader.spectra()

    with ThreadPool(processes=10) as pool:
        data = list(
            tqdm(
                pool.imap(_worker, locations),
                total=len(locations)
            )
        )


    # note that file is a Resource Object here.
    with create_resource(target_file,'wb') as resource:
        # Write the header
        resource.write(source_restart_file.header_bytes())

        # Write the partial spectra sorted by start index of the spectra
        for i_start, partial_spectra in tqdm(
                sorted(data, key=lambda x: x[0]),total=len(data)):

            # Write to file
            resource.write(partial_spectra.tobytes('C'))

        # write the tail
        resource.write(source_restart_file.tail_bytes())


def write_restart_file(spectra:Union[Dataset,numpy.ndarray,DataArray],
                       target_file,
                       parent_restart_file:RestartFile,
                       spectra_are_frequence_energy_density=True
                       ):
    """
    Create a resource file from the given spectra at the target location. The
    target can be a local file or a s3 uri.

    :param spectra: Energy density as funciton of frequency or Action density
        as function of frequency. Needs to have the same number of spectra as
        the parent restart file.
    :param target_file: path to local file or s3 uri to be created.
    :param parent_restart_file: Restart file object of the parent restart file.
        We need this information to create valid header information.
    :param spectra_are_frequence_energy_density: are the spectra energy
        densities. If so we need to transform into wavenumber action density.
        If not, the spectra are assumed to already be valid wavenumber spectra
        and no transformation is applied.
    :return: None
    """
    if isinstance(spectra,Dataset):
        spectra = spectra['variance_density'].values
    elif isinstance(spectra,DataArray):
        spectra = spectra.values


    dtype = numpy.dtype('float32') #
    dtype = dtype.newbyteorder(parent_restart_file._meta_data.byte_order)

    shape = spectra.shape
    if shape[0] != parent_restart_file.number_of_spatial_points:
        raise ValueError('Input spectra have more spatial points than the '
                         'source resource file contains')

    if shape[1] != parent_restart_file.number_of_frequencies:
        raise ValueError('Input spectra have more frequences than the '
                         'source resource file contains')

    if shape[2] != parent_restart_file.number_of_frequencies:
        raise ValueError('Input spectra have more directions than the '
                         'source resource file contains')

    # Ensure we are writing the right floating point accuracy and byte_order
    spectra = spectra.astype(dtype, copy=False)
    if spectra_are_frequence_energy_density:
        conversion_factor = \
            parent_restart_file.to_wavenumber_action_density(slice(None, None, None))
        spectra[:,:,:] = spectra[:,:,:] * conversion_factor[ :,:,None]


    with create_resource(target_file,'wb') as file:
        # Use the parent file to get the valid header.
        file.write(parent_restart_file.header_bytes())
        file.write( spectra.tobytes('C') )
        # Use the parent file to get the "tail", i.e. all the information
        # stored in the restart file after the header. Currently *I* do not
        # know what this information exactly is.
        file.write(parent_restart_file.tail_bytes())


def write_partial_restart_file(
        spectra: Union[Dataset,numpy.ndarray,DataArray],
        target_file: str,
        parent_restart_file: RestartFile,
        s: slice,
        spectra_are_frequence_energy_density=True
        ):
    """
    Create a partial restart file containing only the given spectra and
    their corresponding linear indices in the complete restart file. Used as
    part of a map operation storing the information that was processed in the
    current batch while operating on the total spectra. See
    reassemble_restart_file_from_parts for the reduce operation to create a
    valid restart file from all its parts.

    :param spectra: Energy density as funciton of frequency or Action density
        as function of frequency. Needs to have the same number of spectra as
        the parent restart file.
    :param target_file: path to local file or s3 uri to be created. Needs to
        be unique for the partial files (w.r.t. the other partial files).
    :param parent_restart_file: Restart file object of the parent restart file.
        We need this information to create valid header information.
    :param s: slice- contains the start and end linear_indices. Note that only
        a stepsize of 1 is supported in a slice. And indices *must* be given
        (None is not valid).
    :param spectra_are_frequence_energy_density: are the spectra energy
        densities. If so we need to transform into wavenumber action density.
        If not, the spectra are assumed to already be valid wavenumber spectra
        and no transformation is applied.
    :return:
    """
    if isinstance(spectra,Dataset):
        spectra = spectra['variance_density'].values
    elif isinstance(spectra,DataArray):
        spectra = spectra.values

    dtype = numpy.dtype('float32')  #
    dtype = dtype.newbyteorder(parent_restart_file._meta_data.byte_order)

    shape = spectra.shape
    start, stop, step = s.indices(parent_restart_file.number_of_spatial_points)
    if shape[0] != stop - start:
        raise ValueError('Input spectra have more spatial points than the '
                         'source resource file contains')

    if shape[1] != parent_restart_file.number_of_frequencies:
        raise ValueError('Input spectra have more frequences than the '
                         'source resource file contains')

    if shape[2] != parent_restart_file.number_of_frequencies:
        raise ValueError('Input spectra have more directions than the '
                         'source resource file contains')

    # Ensure we are writing the right floating point accuracy and byte_order
    spectra = spectra.astype(dtype, copy=False)

    if spectra_are_frequence_energy_density:
        spectra[:, :, :] = spectra[:, :, :] * \
                           parent_restart_file.to_wavenumber_action_density(
                               slice(start, stop, step))[:, :, None]

    with create_resource(target_file,'wb') as file:
        location = parent_restart_file.resource.resource_location.encode('utf-8')

        # for a partial file we will just write as a header:
        #  1) the length of the location string (path, uri) of the
        #     source restart file
        #  2) the location string in utf-8
        #  3) the start index in int32, little endian.
        #  4) the stop index in int32, little endian.
        #  5) the spectra themselves.
        # All other information (spectral size etc.) is contained in the
        # source restart file we point to.
        file.write(len(location).to_bytes(4, 'little'))
        file.write(location)
        file.write(start.to_bytes(4, 'little'))
        file.write(stop.to_bytes(4, 'little'))
        file.write(spectra.tobytes('C'))


class _PartialRestartFileReader:
    """
    Helper Class to read partial restart files. A partial restart file is a
    binary file that contains (in order):

        location_length (4 byte int): number of bytes needed for the char array
            that contains the parent file names.
        location (location_length bytes): char array contaning the location of
            the parent restart file (local file path or s3 uri).
        start (4 byte int): start linear index of the arrays
        stop (4 byte int): stop linear index (exclusive) of the arrays.
        spectra: ( start-stop)*numer_of_spectral_points*4 byte float array
            containing the spectral data.
    """
    def __init__(self, location, parent_restart_file):
        """
        :param location: local file path or s3 uri.
        :param parent_restart_file: restart file the partial spectra originate
            from. We need this to be able to write a valid restart file as
            output.
        """
        self.resource = create_resource(location)

        # Read the header information
        self.location_length = int.from_bytes(self.resource.read(4), 'little')
        self.location = self.resource.read(
            self.location_length).decode('utf-8')

        self.start = int.from_bytes(self.resource.read(4), 'little')
        self.stop = int.from_bytes(self.resource.read(4), 'little')

        self.source_restart_file = parent_restart_file
        assert self.source_restart_file.resource.resource_location \
                   == self.location

    def spectra(self) -> numpy.ndarray:
        """
        Read the spectra contained in the partial spectral file and return
        a numpy.ndarray of type float32 with the correct byte-order.
        :return:
        """
        raw_data = self.resource.read()
        dtype = numpy.dtype('float32')
        dtype = dtype.newbyteorder(
            self.source_restart_file._meta_data.byte_order)
        spectra = numpy.frombuffer( raw_data, dtype=dtype )
        spectra = numpy.reshape( spectra,
            (
            self.stop-self.start ,
            self.source_restart_file.number_of_frequencies,
            self.source_restart_file.number_of_directions)
        )
        return spectra


def clone_restart_file(source_restart_file,
                       model_definition_file, target) -> None:
    """
    Clone restart file. Mostly useful to test that the reading and writing
    operations create an exact duplicate of a given restart file (otherwise
    we could just copy the file).

    :param source_restart_file: file to be cloned
    :param model_definition_file: model definition file
    :param target: target file
    :return: None
    """
    source_restart_file = open_restart_file(source_restart_file,
                                            model_definition_file)
    source_restart_file._convert = False
    write_restart_file(source_restart_file[:], target, source_restart_file, False)



# -*- encoding: utf-8 -*-
import setuptools

with open('README.md', 'r') as file:
    readme_contents = file.read()

setuptools.setup(
    name='roguewave',
    version='0.1.62',
    license='Apache 2 License',
    install_requires=[
        'pysofar>=0.1.13',
        'numpy',
        'netCDF4',
        'pandas',
        'qpsolvers',
        'scipy',
        'numba',
        'tqdm',
        'h5netcdf',
        's3fs',
        'boto3',
        'xarray',
        'requests',
        'cfgrib'
    ],
    description='Python package to interact with Sofar wave data',
    long_description=readme_contents,
    long_description_content_type='text/markdown',
    author='Pieter Bart Smit',
    author_email='sofaroceangithubbot@gmail.com',
    url='https://github.com/sofarocean/roguewave.git',

    package_dir={'': 'src'},
    packages=setuptools.find_packages('src'),

    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent"
    ],
    project_urls={
        'Sofar Ocean Site': 'https://www.sofarocean.com'
    }
)



import numpy
from qpsolvers import solve_ls
from .utils import get_direction_increment, get_rhs, get_constraint_matrix

def log_likelyhood(directions_radians: numpy.ndarray, a1, b1, a2,
                     b2) -> numpy.ndarray:
    """
    Return the directional distribution that minimizes the variance (D**2)
    constrained by given observed directional moments,

    :param directions_radians: 1d array of wave directions in radians,
    length[number_of_directions]

    :param a1: 1d array of cosine directional moment as function of frequency,
    length [number_of_frequencies]

    :param b1: 1d array of sine directional moment as function of frequency,
    length [number_of_frequencies]

    :param a2: 1d array of double angle cosine directional moment as function
    of frequency, length [number_of_frequencies]

    :param b2: 1d array of double angle sine directional moment as function of
    frequency, length [number_of_frequencies]

    :return: array with shape [number_of_frequencies,number_of_direction]
    representing the directional distribution of the waves at each frequency.

    Minimize the variance of the solution:

           integrate D**2 over directions

    such that the resulting distribution D reproduces the observed moments.

    Implementation notes:
    - we formulate the problem as a standard Quadratic Programming problem which
      can them be solved efficiently with the qpsolvers package.
    """

    a1 = numpy.atleast_1d(a1)
    b1 = numpy.atleast_1d(b1)
    a2 = numpy.atleast_1d(a2)
    b2 = numpy.atleast_1d(b2)

    number_of_frequencies = len(a1)
    directional_distribution = numpy.zeros(
        (number_of_frequencies, len(directions_radians)))

    constraint_matrix = get_constraint_matrix(directions_radians)
    rhs = get_rhs(a1, b1, a2, b2)
    identity_matrix = numpy.diag(numpy.ones_like(directions_radians), 0)

    zeros = numpy.zeros_like(directions_radians)
    direction_increment = get_direction_increment(directions_radians)
    upperbound = numpy.ones_like(
        directions_radians) / direction_increment.min()

    for ifreq in range(0, number_of_frequencies):
        res = solve_ls(R=identity_matrix, s=zeros,
                       # minimizing |Rx-b|**2
                       lb=zeros, ub=upperbound,
                       # lb: non-negative; ub: binwidth * ub = 1
                       A=constraint_matrix, b=rhs[ifreq, :], verbose=False
                       # with hard constraint that Ax=b
                       )

        if res is None:
            raise Exception('No solution')

        directional_distribution[ifreq, :] = res

    directional_distribution[directional_distribution < 0] = 0
    return numpy.squeeze(directional_distribution)


"""

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Smit
"""
import os
import dotenv
import requests
import json
from pysofar import SofarConnection
from pysofar.wavefleet_exceptions import QueryError
from functools import cached_property
from roguewave.wavespectra import FrequencyDirectionSpectrum, \
    create_2d_spectrum
import typing
import netCDF4
import numpy
from typing import List, overload, Union, Dict
from roguewave.tools.time import to_datetime64


def get_token():
    # config values
    userpath = os.path.expanduser("~")
    enviromentFile = os.path.join(userpath, 'sofar_api.env')
    dotenv.load_dotenv(enviromentFile)
    token = os.getenv('SPECTRAL_API_TOKEN')
    _wavefleet_token = token

    return _wavefleet_token


def get_endpoint():
    _endpoint = os.getenv('SPECTRAL_URL')
    if _endpoint is None:
        _endpoint = 'https://api.sofarocean.com/api'
    return _endpoint


class SofarSpectralAPI(SofarConnection):
    def __init__(self, costum_token=None):
        #
        token = costum_token if costum_token else get_token()
        super().__init__(token)
        self.endpoint = get_endpoint()

    @cached_property
    def _query(self):
        scode, data = self._get('op-wave-spectra')

        if scode != 200:
            raise QueryError(data['message'])

        return data['data']

    def _download(self, latitude, longitude):
        # Helper methods to download files

        url = f"{self.endpoint}/op-wave-spectra/{latitude}/{longitude}"
        response = requests.get(url, headers=self.header)

        status = response.status_code
        if status != 200:
            data = json.loads(response.text)
            raise QueryError(data['message'])

        data = response.content
        return data

    def points(self):
        data = self._query
        return [{'latitude': x['latitude'], 'longitude': x['longitude']} for x
                in data]

    def urls(self):
        data = self._query
        return [x['url'] for x in data]

    def download_spectral_file(self, latitude, longitude, directory) -> str:
        os.makedirs(directory, exist_ok=True)
        filename = f'spectrum_{longitude}E_{latitude}N.nc'
        filepath = os.path.join(directory, filename)

        with open(filepath, 'wb') as file:
            file.write(self._download(latitude, longitude))

        return filepath

    def download_all_spectral_files(self, directory) -> typing.List[str]:
        points = self.points()
        return [
            self.download_spectral_file(point['latitude'], point['longitude'],
                                        directory) for point in points]


@overload
def load_sofar_spectral_file(filename: str)->FrequencyDirectionSpectrum: ...

@overload
def load_sofar_spectral_file(filename: List[str])->List[FrequencyDirectionSpectrum]: ...

@overload
def load_sofar_spectral_file(filename: Dict[str,str])->Dict[str,FrequencyDirectionSpectrum]: ...

def load_sofar_spectral_file(
        filename: Union[str, List[str], Dict[str, str]]):

    if isinstance(filename,str):
        if not os.path.exists(filename):
            raise FileNotFoundError()

        dataset = netCDF4.Dataset(filename)

        nt = dataset.dimensions['time'].size
        frequencies = dataset.variables['frequencies'][:]
        directions = dataset.variables['directions'][:]

        spectra = []
        for ii in range(0, nt):
            unixepochtime = dataset.variables['time'][ii]
            latitude = dataset.variables['latitude'][0]
            longitude = dataset.variables['longitude'][0]
            variance_density = \
                numpy.squeeze(
                    dataset.variables['frequency_direction_spectrum'][ii, :, :])

            if dataset.variables[
                'frequency_direction_spectrum'].dimensions[1] == 'directions':
                variance_density = variance_density.transpose()

            spectra.append(
                create_2d_spectrum(
                    frequency=frequencies,
                    variance_density=variance_density,
                    time=to_datetime64(unixepochtime),
                    latitude=latitude,
                    longitude=longitude,
                    direction=directions,
                )
            )
        return FrequencyDirectionSpectrum.concat_from_list(spectra)
    elif isinstance(filename,dict):
        out = {}
        for key in filename:
            out[key] = load_sofar_spectral_file(filename[key])
        return out
    elif isinstance(filename,List):
        return [ load_sofar_spectral_file(name) for name in filename]
    else:
        raise Exception('Unsupported input type for filename')





class ExceptionCouldNotDownloadData(Exception):
    """
    Query raised when no frequency data is available for the spotter
    """
    pass


class ExceptionNoDataForVariable(Exception):
    """
    Query raised when no frequency data is available for the spotter
    """
    pass



"""
Contents: Simple file caching routines that automatically cache remote files
          locally for use.

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Classes:
- `FileCache`, main class implementing the Caching structure. Should not
   directly be invoked. Instead, fetching/cache creation is controlled by a
   set of function defined below

Functions:

"""
import hashlib
import os
from _warnings import warn
from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import Union, List, Tuple, Callable, Dict
from dataclasses import dataclass
from tqdm import tqdm
from roguewave import logger
from .remote_resources import RemoteResourceS3, \
    RemoteResourceHTTPS, RemoteResource, RemoteResourceLocal
from .exceptions import _RemoteResourceUriNotFound
from json import dumps, load

TEMPORARY_DIRECTORY = '~/temporary_roguewave_files/filecache/'
CACHE_SIZE_GB = 5
MAXIMUM_NUMBER_OF_WORKERS = 10
KILOBYTE = 1000
MEGABYTE = 1000 * KILOBYTE
GIGABYTE = 1000 * MEGABYTE


@dataclass()
class CacheMiss:
    uri: str
    filepath: str
    filename: str
    allow_for_missing_files: bool
    post_process_function: Callable[[str], None]
    download_function: Callable[[str, str], bool] = None


class FileCache:
    """
    Simple file caching class that when given an URI locally stores the
    file in the cache directory and returns the path to the file. The file
    remains in storage until the cache directory exceeds a prescribed size,
    in which case files with oldest access/modified dates get deleted first
    until everything fits in the cache again. Any time a file is accessed it's
    modified date gets updated so that often used files automaticall remain in
    cache.

    The files are stored locally in the directory specified on class
    initialization, as:

        [path]/CACHE_PREFIX + md5_hash_of_URI + CACHE_POSTFIX

    The pre- and post fix are added so we have an easy pattern to distinguish
    cache files from other files.

    Methods
      * __getitem__(keys) : accept a simgle uri_key or a list of URI's and
        returns filepaths to local copies thereof. You would typically use the
            cache[keys] notation instead of the dunder method.
      * purge() clear all contents of the cache (destructive, deletes all local
        files).

    Usage:

        cache = FileCache()
        list_of_local_file_names = cache[ [list of URI's ] ]

    # do stuff with file
    ...
    """

    CACHE_FILE_PREFIX = 'cachefile_'
    CACHE_FILE_POSTFIX = '_cachefile'

    def __init__(self,
                 path: str = TEMPORARY_DIRECTORY,
                 size_GB: Union[float, int] = CACHE_SIZE_GB,
                 do_cache_eviction_on_startup: bool = False,
                 resources: List[RemoteResource] = None,
                 parallel=True,
                 allow_for_missing_files=False,
                 ):
        """
        Initialize Cache
        :param path: path to store cache. If path does not exist it will be
            created.
        :param size_GB: Maximum size of the cache in GiB. If cache exceeds
            the size, then files with oldest access/modified dates get deleted
            until everthing fits in the cache again. Fractional values (floats)
            are allowed.
        :param do_cache_eviction_on_startup: whether we ensure the cache size
            conforms to the given size on startup. If set to true, a cache
            directory that exceeds the maximum size will be reduced to max
            size. Set to False by default in which case an error occurs. The
            latter to prevent eroneously evicting files from a cache that was
            previously created on purpose with a larger size that the limit.


        """
        self.path = os.path.expanduser(path)
        # create the path if it does not exist
        if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)

        self.config = {
            "size_gb": size_GB,
            "parallel": parallel,
            "allow_for_missing_files":allow_for_missing_files
        }
        if self.config_exists():
            self.config |= self.load_config()
        else:
            self._write_config()

        # Some counters to keep track of total cache misses, hits and
        # evictions. No downstream use right now/
        self._cache_misses = 0
        self._cache_hits = 0
        self._cache_evictions = 0
        self.disable_progress_bar = False

        # initialize the cache.
        self._entries = {}  # the key/value pair cache
        self._initialize_cache(do_cache_eviction_on_startup)

        # Post processing and validation functions
        self.directives = {"validate": {}, "postprocess":{}}

        # download resources
        if resources is None:
            self.resources = [RemoteResourceS3(),
                              RemoteResourceHTTPS(),
                              RemoteResourceLocal()]
        else:
            self.resources = resources

    @property
    def config_name(self) -> str:
        return os.path.join(self.path,'file_cache_config.json')

    def config_exists(self)->bool:
        return os.path.exists(self.config_name)

    def load_config(self)->Dict:
        with open( self.config_name, 'rb' ) as fp:
            return load(fp)

    def _update_config(self, key, value, write=True):
        self.config[key] = value
        if write:
            self._write_config()

    def _write_config(self):
        with open( os.path.join(
                self.path,'file_cache_config.json'),'wt' ) as fp:
            fp.write(dumps(self.config,indent=4))

    @property
    def max_size(self)->int:
        return self.config['size_gb']

    @max_size.setter
    def max_size(self,size_gb:float):
        self._update_config('size_gb',size_gb)

    @property
    def max_size_bytes(self)->int:
        return  int( self.config['size_gb'] * GIGABYTE)

    @max_size_bytes.setter
    def max_size_bytes(self,size_bytes:int):
        self._update_config('size_gb',size_bytes / GIGABYTE)

    @property
    def parallel(self)->bool:
        return self.config['parallel']

    @parallel.setter
    def parallel(self, parallel:bool):
        self._update_config('parallel',parallel)

    @property
    def allow_for_missing_files(self)->bool:
        return self.config['allow_for_missing_files']

    @allow_for_missing_files.setter
    def allow_for_missing_files(self, allow_for_missing_files:bool):
        self._update_config('allow_for_missing_files',allow_for_missing_files)

    def set_directive_function(
            self,
            directive,
            name,
            function: Union[
                Callable[[str],None],
                Callable[[str],bool]
            ]
        ):

        if directive not in self.directives:
            raise KeyError(f'{directive} is not a valid cache directive.')

        if name in self.directives[directive]:
            raise ValueError(f'Function  for {name} already exists')
        else:
            self.directives[directive][name] = function


    def remove_directive_function(self,directive:str,name:str):

        if directive not in self.directives:
            raise KeyError(f'{directive} is not a valid cache directive.')

        if name not in self.directives[directive]:
            raise ValueError(f'Function  for {name} does not exist')
        else:
            self.directives[directive].pop(name)


    def _cache_file_name(self, uri: str) -> str:
        """
        Return the filename that corresponds to the given uri. We construct
        the file name using a simple md5 hash of the uri string prefixed
        with a cache file prefix. THe later is introduced to seperate cache
        files in a path from user files (and avoid including/deleting those).

        :param uri: valid uri stripped from directives
        :return: valid cache file
        """
        return self.CACHE_FILE_PREFIX + _hashname(
            uri) + self.CACHE_FILE_POSTFIX

    def _cache_file_path(self, uri: str) -> str:
        """
        Construct the path where the given uri is stored locally.

        :param uri: valid uri stripped from directives.
        :return: valid cache file
        """
        return os.path.join(self.path, self._cache_file_name(uri))

    def _get_cache_files(self) -> List[str]:
        """
        Find all files that are currently a member of the cache.
        :return:
        """
        _cache_files = []
        for path, dirs, files in os.walk(self.path):
            # Return all files that are "cache" objects. This is a safety if
            # other user files are present, so that these don't accidentally
            # evicted from the cache (aka deleted).
            return [file for file in files if
                    file.startswith(self.CACHE_FILE_PREFIX) and
                    file.endswith(self.CACHE_FILE_POSTFIX)]
        else:
            return []

    def _initialize_cache(self, do_cache_eviction_on_startup: bool) -> None:
        """
        Initialize the file cache. Look on disk for files in the cache path
        that have the required prefix and load these into the cache. Once
        loaded, we do a check whether or not the cache is full and if we need
        to remove files.
        :param do_cache_eviction_on_startup: see description under __init__
        :return:
        """
        self._entries = {}
        for file in self._get_cache_files():
            filepath = os.path.join(self.path, file)
            self._entries[file] = filepath

        # See if cache is still < size
        if do_cache_eviction_on_startup:
            self._cache_eviction()
        else:
            if self._size() > self.max_size_bytes:
                raise ValueError('The cache currently existing on disk '
                                 'exceeds the maximum cache size of the '
                                 f'current cache ({self.max_size} gb).'
                                 f'\n The cache size can be increased by'
                                 f' editting the cache config file: '
                                 f'{self.config_name}')

    def in_cache(self, unparsed_uris) -> List[bool]:
        # make sure input is a list
        if isinstance(unparsed_uris, str):
            unparsed_uris = [unparsed_uris]

        uris, _ = parse_directives(unparsed_uris)

        # Create the hashes from the URI's
        hashes = [self._cache_file_name(uri) for uri in uris]
        return [self._is_in_cache(_hash) for _hash in hashes]

    def _is_in_cache(self, _hash: str) -> bool:
        """
        Check if a _hash is in the cache
        :param _hash: hash to check
        :return: True if in Cache, False if not.
        """
        cache_hit = _hash in self._entries
        return cache_hit

    def _add_to_cache(self, _hash: str, filepath: str) -> None:
        # add entry to the cache.
        self._entries[_hash] = filepath

    def remove(self, unparsed_uri: str) -> None:
        """
        Remove an entry from the cache
        :param unparsed_uri: uri
        :return: None
        """
        uri, _ = parse_directive(unparsed_uri)

        if not self.in_cache(uri):
            raise ValueError(f'Key {uri} not in Cache')

        _hash = self._cache_file_name(uri)
        return self._remove_item_from_cache(_hash)

    def _remove_item_from_cache(self, _hash: str) -> None:
        """
        Remove a hash key from the cache. Here it is assumed that the _hash is
        a valid entry. We do allow for non existance of corresponding files as
        the cache can get out of sync if something external deleted the file.
        Since the endstate is valid (no entry in cache, no entry on disk) this
        is considered OK.

        :param _hash: hash key
        :return: None
        """

        assert _hash in self._entries

        file_to_delete = self._entries.pop(_hash)

        if os.path.exists(file_to_delete):
            logger.debug(f' - removing {_hash}')

            # And delete file.
            os.remove(file_to_delete)
        else:
            logger.debug(f' - file {_hash} did not exist on disk')

        return None

    def _get_from_cache(self, _hash: str) -> str:
        """
        Get entry from cache and touch the file to indicate it has been used
        recently (last to be evicted)
        :param _hash: file_hash corresponding to uri
        :return: file path
        """
        filepath = self._entries[_hash]

        if not os.path.exists(filepath):
            raise FileNotFoundError('The filepath in the cache log does not'
                                    'exist on disk.')

        # Touch the file to indicate we recently used it.
        Path(filepath).touch()

        return filepath

    def get_cache_misses(self, uris: List[str],
                         directives: List[Dict[str, str]]) \
            -> List[CacheMiss]:
        """
        Function to get all cache misses and return a list of CacheMiss objects
        needed to download the misses from remote resources.

        This function also perform validates on potential cache hits if a
        relevant validation function is set *and* validation is requested
        through a directive.

        :param uris: list of uris stripped of directives
        :param directives: list of directives per uri (empty dict if none)
        :return: list of cache misses
        """

        cache_misses = []
        for uri, directive in zip(uris, directives):

            # what is the hashkey/filename
            hashkey = self._cache_file_name(uri)
            filepath = self._cache_file_path(uri)

            # is the key in cache?
            valid_entry = False
            if self._is_in_cache(hashkey):
                # If so is it a valid entry
                if 'validate' in directive:
                    # Call the user supplied validation function with the
                    # filepath as argument
                    validation_function = \
                        self.directives['validate'][directive['validate']]
                    valid_entry = validation_function(filepath)

                    if not valid_entry:
                        # remove the locally stored entry if not valid
                        os.remove(filepath)
                else:
                    # Defaults to True if no validation directive is given
                    valid_entry = True

            if not valid_entry:
                # If not a valid entry (either missing or invalid)
                #
                if 'postprocess' in directive:
                    # Add the postprocess function to use if requested.
                    post_process_function = \
                        self.directives[
                            'postprocess'][directive['postprocess']]

                else:
                    # otherwise set a null function as postprocessor
                    post_process_function = do_nothing

                cache_misses.append(
                    CacheMiss(
                        uri=uri,
                        filepath=filepath,
                        filename=hashkey,
                        allow_for_missing_files=self.allow_for_missing_files,
                        post_process_function=post_process_function
                    )
                )
        return cache_misses

    def __len__(self) -> int:
        """
        :return: Number of entries in the cache.
        """
        return len(self._entries)

    def __getitem__(self, unparsed_uris: Union[List, str]) -> List[str]:
        """
        Get filenames corresponding to locally stored versions of the objects
        the URI points to. Note that the unparsed_uris take the form:

        [ directive=option ; ... directive=option ] ":" [scheme] "://" [path]

        e.g for amazon s3 where we want to perform validation and post
            processing on entries:

            validate=grib;postprocess=grib:s3://bucket/key

        or without cache directives

            s3://bucket/key

        Cache directives are optional, but if specified the corresponding
        user defined handling function must have been set. By default no
        validation or postprocessing functions are set.

        :param unparsed_uris: URI's that may still include directives.
        :return:
        """
        # make sure input is a list
        if isinstance(unparsed_uris, str):
            unparsed_uris = [unparsed_uris]

        # Remove cache directives from uris (if included)
        uris, directives = parse_directives(unparsed_uris)
        filepaths = [self._cache_file_path(uri) for uri in uris]

        # for all URI's not in cache
        if cache_misses := self.get_cache_misses(uris, directives):
            _ = _download_from_resources(
                cache_misses,
                self.resources,
                parallel_download=self.parallel,
                disable_progress_bar=self.disable_progress_bar
            )

            for cache_miss in cache_misses:
                self._add_to_cache(cache_miss.filename,
                                   cache_miss.filepath)

        size_of_requested_data = _get_total_size_of_files_in_bytes(filepaths)
        if size_of_requested_data > self.max_size_bytes:
            warning = f'The requested data does not fit into the cache.' \
                      f'To avoid issues the cache is enlarged to ensure' \
                      f'the current set of files fits in the cache. \n' \
                      f'old size: {self.max_size_bytes} bytes; ' \
                      f'new size {size_of_requested_data + MEGABYTE}'
            warn(warning)
            logger.warning(warning)
            self.max_size_bytes = size_of_requested_data + MEGABYTE

        self._cache_misses += len(cache_misses)
        self._cache_hits += len(uris) - len(cache_misses)

        # See if we need to do any cache eviction because the cache has become
        # to big.
        if not len(cache_misses) == 0:
            self._cache_eviction()

        return filepaths

    def _cache_eviction(self) -> bool:
        """
        Simple cache eviction policy. If the cache exceeds the maximum size
        remove data from the cache based on whichever file was interacted with
        the longest time ago. Evict files until we are below the acceptable
        cache size.

        :return: True if eviction occured, False otherwise.
        """

        # check if we exceed the size, if not return
        if not self._size() > self.max_size_bytes:
            return False

        # Get access/modified times for all the files in cache
        modified = []
        for _hash, fp in self._entries.items():
            # From my brief reading, access time is not always reliable,
            # hence I use whatever the latest time set is for modified or
            # access time as an indicator of when we last interacted with
            # the file.
            access_time = os.path.getatime(fp)
            modified_time = os.path.getmtime(fp)

            # pick whichever is most recent.
            time_to_check = access_time if access_time > modified_time \
                else modified_time
            modified.append((time_to_check, _hash))

        # Sort files in reversed chronological order.
        files_in_cache = [x[1] for x in
                          sorted(modified, key=lambda x: x[0], reverse=True)]

        # Delete files one by one as long as the cache_size exceeds the max
        # size.
        while (_size := self._size()) > self.max_size_bytes:
            self._cache_evictions += 1
            logger.debug(
                f'Cache exceeds limits: {_size} bytes, max size: '
                f'{self.max_size_bytes} bytes')

            # Get the hash and path of the oldest file and remove
            self._remove_item_from_cache(files_in_cache.pop())

        return True

    def _size(self) -> int:
        """
        Return size on disk of the cache in bytes.
        :return: cache size in bytes.
        """
        return _get_total_size_of_files_in_bytes(
            list(self._entries.values()), self.path)

    def purge(self) -> None:
        """
        Delete all the files in the cache.
        :return: None
        """
        logger.debug(f'Purging cache')
        keys = list(self._entries.keys())
        for key in keys:
            filepath = self._entries.pop(key)
            logger.debug(f' - deleting {filepath}')
            os.remove(filepath)
        logger.debug(f'Purging cache done')


def _download_from_resources(cache_misses: List[CacheMiss],
                             resources: List[RemoteResource],
                             parallel_download=False,
                             disable_progress_bar = False,
                             ) -> List[bool]:
    """
    Wrapper function to download multiple uris from the resource(s).
    :param cache_misses: List containing cache misses to download
    :param parallel_download: If true, downloading is performed in parallel.
    :return: List of boolean indicating if the download was a success.
    """

    def _worker(cache_miss: CacheMiss) -> bool:
        try:
            cache_miss.download_function(cache_miss.uri, cache_miss.filepath)
            cache_miss.post_process_function(cache_miss.filepath)
            return True
        except _RemoteResourceUriNotFound as e:
            if cache_miss.allow_for_missing_files:
                warning = f'Uri not retrieved: {str(e)}'
                warn(warning)
                logger.warning(warning)
            else:
                raise e
            return False

    # construct the arguments to be used for parallel downloading of files.
    # Specifically, we need to match the right resource for downloading to the
    # right URI.
    for cache_miss in cache_misses:
        # Loop over all resources until we find one that can interpret the URI
        # (this is pretty naive approach and should probably be refactored to
        #  some direct mapping if the number of resources ever gets very long)
        for resource in resources:
            # For each resource check if the resource can interpret the URI
            if resource.valid_uri(cache_miss.uri):
                # If so, get the download function, and other arguments and
                # break
                cache_miss.download_function = resource.download()
                break
        else:
            # If we didn't break the loop no valid resource was found, raise
            # error
            raise ValueError(f'No resource available for URI: '
                             f'{cache_miss.uri}')

    # Download the requested objects.
    if parallel_download and len(cache_misses) > 1:
        with ThreadPool(processes=MAXIMUM_NUMBER_OF_WORKERS) as pool:
            output = list(
                tqdm(
                    pool.imap(_worker, cache_misses),
                    total=len(cache_misses)
                )
            )
    else:
        if len(cache_misses) == 1:
            disable_progress_bar = True

        output = list(
            tqdm(
                map(_worker, cache_misses),
                total=len(cache_misses),disable=disable_progress_bar
            )
        )
    return output


def _get_total_size_of_files_in_bytes(filenames: List[str], path=None) -> int:
    """
    Simple function to calculate the size of a list of files on disk.
    :param filenames: list of filenames or filepaths
    :param path: if filenames are provided, this lists the path, otherwise set
        to None

    :return: Total size in bytes
    """
    size = 0
    for filename in filenames:
        if path is None:
            filepath = filename
        else:
            filepath = os.path.join(path, filename)
        size += os.path.getsize(filepath)
    return size


def _hashname(string: str) -> str:
    """
    Returns a md5 hash of a given string.
    :param string: input string
    :return: hexdigest of md5 hash.
    """
    return hashlib.md5(string.encode(), usedforsecurity=False).hexdigest()


def parse_directives(raw_uris: List[str]) -> Tuple[List[str], List[dict]]:
    uris = []
    directives = []
    for raw_uri in raw_uris:
        uri, directive = parse_directive(raw_uri)
        uris.append(uri)
        directives.append(directive)
    return uris, directives


def parse_directive(unparsed_uri: str) -> Tuple[str, dict]:
    """
    unparsed_uris take the form:

        [ directive=option ; ... directive=option ] ":" [scheme] "://" [path]

        e.g for amazon s3 where we want to perform validation and post
            processing on entries:

            validate=grib;postprocess=grib:s3://bucket/key

        or without cache directives

            s3://bucket/key

    This function seperates the directive/option pairs into a directove
    dictionary, and a valid uri, i.e.

                validate=grib;postprocess=grib:s3://bucket/key

    becomes

        directive = { "validate":"grib", "postprocess":"grib}
        uri = s3://bucket/key

    The parsing is really simple.

    :param unparsed_uri: uri possibly containing cache directives
    :return:
    """

    # split in directives_scheme part and a path.
    directives_and_scheme, path = unparsed_uri.split('://')

    parsed_directives = {}
    # if a colon is present then directives are provided.
    if ':' in directives_and_scheme:
        # split directives from the scheme
        directives, scheme = directives_and_scheme.split(':')

        # split multiple directives (if present)
        directives = directives.split(';')

        # for each directive store in the dict.
        for directive in directives:
            directive_name, directive_parameter = directive.split('=')
            parsed_directives[directive_name] = directive_parameter
    else:
        # no directives
        scheme = directives_and_scheme

    uri = scheme + '://' + path
    return uri, parsed_directives


def do_nothing(*arg, **kwargs):
    """
    Null function for convenience
    :param arg:
    :param kwargs:
    :return:
    """
    return None



"""
Contents: Routines to get data from the spotter api

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Routines to calculate times we have output at different init and forecasthours

Functions:

- `find_clostest_init_time`, find closest model init time to a given valid time
- `timebase forecast`, get valid forecast times
- `timebase_lead`, get all output from the model at a given lead time.
- `timebase_eval`, get all output from different forecasts at a given
    evaluation time.
"""


# Import
# =============================================================================

from datetime import datetime, timezone, timedelta
from typing import List, Union, Tuple
from roguewave.tools.time import to_datetime_utc


# Model private variables
# =============================================================================


INTERVAL_TYPE = Union[
    List[tuple[int, timedelta]], Tuple[tuple[int, timedelta]]
]

# Classes
# =============================================================================


class ModelTimeConfiguration:
    def __init__(self,
                 cycle_time_hours: timedelta = timedelta(hours=6),
                 cycle_offset_hours: timedelta = timedelta(hours=0),
                 output_interval: INTERVAL_TYPE = (
                         (239, timedelta(hours=1)),)):

        self.cycle_time_hours = cycle_time_hours
        self.cycle_offset_hours = cycle_offset_hours
        self.output_interval = output_interval
        self.duration = self.forecast_hours()[-1]

    @property
    def cycle_time_hours(self):
        return self._cycle_time_hours

    @cycle_time_hours.setter
    def cycle_time_hours(self, value):
        self._cycle_time_hours = _to_timedelta(value)

    @property
    def cycle_offset_hours(self):
        return self._cycle_offset_hours

    @cycle_offset_hours.setter
    def cycle_offset_hours(self, value):
        self._cycle_offset_hours = _to_timedelta(value)

    @property
    def output_interval(self) -> INTERVAL_TYPE:
        return self._output_interval

    @output_interval.setter
    def output_interval(self, values):
        self._output_interval = []
        for value in values:
            if isinstance(value[1], timedelta):
                self._output_interval.append(value)
            else:
                self._output_interval.append(
                    (value[0], timedelta(hours=value[1])))

    def forecast_hours(self, duration: timedelta = None) -> List[timedelta]:
        forecast_hours = [timedelta(hours=0)]
        for interval in self.output_interval:
            base_hour = forecast_hours[-1]
            for index in range(interval[0]):
                forecast_hours.append(
                    base_hour + interval[1] * (index + 1)
                )
                if duration is not None:
                    if forecast_hours[-1] >= duration:
                        return forecast_hours
        return forecast_hours

    def interval_at_lead_time(self, lead_time: timedelta) -> timedelta:
        """
        Get the output interval at a certain lead time
        :param lead_time:
        :return: output interval at the given lead time
        """
        if lead_time > self.duration:
            raise ValueError('Lead time exceeds duration of the forecast')

        forecast_hours = [timedelta(hours=0)]
        for interval in self.output_interval:
            base_hour = forecast_hours[-1]
            if lead_time <= base_hour + interval[0] * interval[1]:
                interval = interval[1]
                break
        else:
            interval = self.output_interval[-1][1]
        return interval


class TimeSlice:
    def __init__(self, start_time:datetime, end_time:datetime):
        self.start_time = to_datetime_utc(start_time)
        self.end_time = to_datetime_utc(end_time)

    def time_base(self, time_configuration:ModelTimeConfiguration)\
            -> List[Tuple[datetime, timedelta]]:
        pass


class TimeSliceForecast(TimeSlice):
    def __init__(self, init_time:datetime, duration:timedelta):
        """
        Timeslice for a single forecast
        :param init_time: init time of forecast (datetime)
        :param duration: forecast length (time delta)
        """
        self.init_time = to_datetime_utc(init_time)
        self.duration = duration
        super().__init__(self.init_time,self.init_time+duration)

    def time_base(self, time_configuration:ModelTimeConfiguration):
        return timebase_forecast(
            init_time=self.init_time,
            duration=self.duration,
            time_configuration=time_configuration)


class TimeSliceLead(TimeSlice):
    def __init__(self,start_time:datetime,end_time:datetime,
                 lead_time:timedelta,exact=False):
        """
        Slice at constant lead time
        :param start_time: start time of period of interest (datetime)
        :param end_time: end time of period of interest (datetime)
        :param lead_time: lead time (timedelta) of interest.
        :param exact: boolean. Do we want data only exactly at the given lead
            time, or as close as possible
        """
        super().__init__(start_time,end_time)
        self.lead_time = lead_time
        self.exact = exact

    def time_base(self, time_configuration:ModelTimeConfiguration):
        return timebase_lead(
            start_time = self.start_time,
            end_time = self.end_time,
            lead_time=self.lead_time,
            time_configuration=time_configuration,
            exact=self.exact)


class TimeSliceAnalysis(TimeSlice):
    def __init__(self,start_time:datetime,end_time:datetime):
        """
        Analysis time
        :param start_time: start time of period of interest (datetime)
        :param end_time: end time of period of interest (datetime)
        """
        super().__init__(start_time,end_time)

    def time_base(self, time_configuration:ModelTimeConfiguration):
        return timebase_lead(
            start_time = self.start_time,
            end_time = self.end_time,
            lead_time=timedelta(hours=0),
            time_configuration=time_configuration,
            exact=False)


class TimeSliceBestForecast(TimeSliceAnalysis):
    def __init__(self,start_time:datetime,end_time:datetime):
        """
        Analysis time
        :param start_time: start time of period of interest (datetime)
        :param end_time: end time of period of interest (datetime)
        """
        super().__init__(start_time,end_time)

    def time_base(self, time_configuration:ModelTimeConfiguration):
        return super().time_base(time_configuration)

class TimeSliceEvaluation(TimeSlice):
    def __init__(self, evaluation_time:datetime, maximum_lead_time:timedelta):
        """
        All data for a constant point in time
        :param evaluation_time: evaluation point (datetime)
        :param maximum_lead_time:  maximum lead time considered (timedelta)
        """
        super().__init__(start_time=evaluation_time,end_time=evaluation_time)
        self.evaluation_time = evaluation_time
        self.maximum_lead_time = maximum_lead_time

    def time_base(self, time_configuration:ModelTimeConfiguration):
        return timebase_evaluation(
            evaluation_time=self.evaluation_time,
            maximum_lead_time=self.maximum_lead_time,
            time_configuration=time_configuration)


# Main Public Functions
# =============================================================================
def find_closet_init_time(evaluation_time: datetime,
                          cycle_time_hours: timedelta = timedelta(hours=6),
                          cycle_offset_hours: timedelta = timedelta(hours=0),
                          lead_time: timedelta = timedelta(hours=0)
                          ) -> datetime:
    """
    Find the model init time that is closest to the evaluation time requested
    so that init_time <= evaluation_time - lead_time

    :param evaluation_time: Time of interest
    :param cycle_time_hours:  cycle time of the model
    :param cycle_offset_hours: offset of the model
    :param lead_time: Lead time of interest
    :return: init time
    """
    evaluation_time_seconds = evaluation_time.timestamp()
    cycle_time_seconds = cycle_time_hours.total_seconds()
    cycle_offset_seconds = cycle_offset_hours.total_seconds()
    lead_time_seconds = lead_time.total_seconds()

    # Find the nearest cycle number.
    number_of_cycles = (
            (evaluation_time_seconds - cycle_offset_seconds
             - lead_time_seconds) // cycle_time_seconds
    )

    init_time = datetime.fromtimestamp(
        number_of_cycles * cycle_time_seconds + cycle_offset_seconds,
        tz=timezone.utc
    )

    # The delta represents the actual lead time of the evaluation time
    # compared to the init time returned
    delta = evaluation_time_seconds - lead_time_seconds - init_time.timestamp()
    if delta >= cycle_time_seconds / 2:
        # If the delta is larger than half a cycle, a cycle with a longer lead
        # time is actually closer to the desired lead time. For example if
        # the requested lead_time is 6 hours, the cycle time 6 hours and the
        # actual lead time for init_time t is one hours, it makes sense to
        # return t0 - 6 hours as the init time as that has a lead time of 7
        # hours, which is closer to the desired lead time of 6 hours, and
        # consequently more representative of the desired forecast accuracy.
        new_lead_time = evaluation_time - init_time - cycle_time_hours
        if new_lead_time.total_seconds() >= 0:
            init_time = init_time + cycle_time_hours

    return init_time


def timebase_forecast(init_time: datetime, duration: timedelta,
                      time_configuration: ModelTimeConfiguration
                      ) -> List[Tuple[datetime, timedelta]]:
    """
    Get all the valid times for a given forecast up to the requested duration.
    Valid times are returned as a List of pairs: (init_time, forecast_hour)

    :param init_time: init time of the forecast
    :param duration: Duration requested
    :param time_configuration:  Output time configuration of the model
    :return: List of (init_time, forecast_hour) so that valid_time =
        init_time + forecast_hour
    """

    return [(init_time, forecast_hour) for forecast_hour in
            time_configuration.forecast_hours(duration=duration)]


def timebase_lead(start_time: datetime, end_time: datetime,
                  lead_time: timedelta,
                  time_configuration: ModelTimeConfiguration, exact=True
                  ) -> List[Tuple[datetime, timedelta]]:
    """
    Get all the valid times at a given lead-time in the period of interest.
    Valid times are returned as a List of pairs: (init_time, forecast_hour).
    If exact is true, only valid-times at exactly the requested lead-time are
    returned. Otherwise, it includes close valid_times as well.

    :param start_time: start time of interval of interest
    :param end_time: end time of interval of interest
    :param lead_time:  lead time of interest
    :param time_configuration: Output time configuration of the model
    :param exact: only return valid times that exactly match the requested lead
        time.
    :return: List of (init_time, forecast_hour) so that valid_time =
        init_time + forecast_hour
    """

    # Make sure the lead time does not exceed the duration
    if lead_time > time_configuration.duration:
        raise ValueError(f'Lead time {lead_time} exceeds maximum forecast '
                         f'length of {time_configuration.duration}')

    forecast_hours = time_configuration.forecast_hours()

    # If exact, make sure the lead time is part of the forecast.
    if exact and (lead_time not in forecast_hours):
        raise ValueError(f'Forecast does not have output '
                         f'exactly at lead time: {lead_time}')

    # Get the output interval at the requested lead time
    interval = time_configuration.interval_at_lead_time(lead_time=lead_time)

    # Calculate number of instances we will return (upper bound)
    n = int((end_time - start_time) / interval)

    timebase = []
    for index in range(0, n):
        #
        valid_time = start_time + index * \
                     time_configuration.output_interval[0][1]

        init_time = find_closet_init_time(
            valid_time,
            cycle_time_hours=time_configuration.cycle_time_hours,
            cycle_offset_hours=time_configuration.cycle_offset_hours,
            lead_time=lead_time)

        forecast_hour = valid_time - init_time

        # if exact=False, and we are on the edge between two different output
        # intervals not all of the forecast hours returned may exist.
        if forecast_hour not in forecast_hours:
            continue

        # If exact, only add at exact lead times.
        if exact and not (forecast_hour == lead_time):
            continue

        timebase.append((init_time, forecast_hour))

    return timebase


def timebase_evaluation(evaluation_time: datetime,
                        maximum_lead_time: timedelta,
                        time_configuration: ModelTimeConfiguration
                        ) -> List[Tuple[datetime, timedelta]]:
    """
    Find all init times and forecast hours so that
        init_time + forecast = evaluation_time,

    :param evaluation_time: Time we want to compare different forecasts for
    :param maximum_lead_time: Maximum lead time of the forecasts to consider
    :param time_configuration: Output time configuration of the model.
    :return: List of (init_time, forecast_hour) so that evaluation_time =
        init_time + forecast_hour
    """

    forecast_hours = time_configuration.forecast_hours(
        duration=maximum_lead_time)

    timebase = []
    for forecast_hour in forecast_hours:
        init_time = find_closet_init_time(
            evaluation_time=evaluation_time,
            cycle_time_hours=time_configuration.cycle_time_hours,
            cycle_offset_hours=time_configuration.cycle_offset_hours,
            lead_time=forecast_hour
        )
        if evaluation_time == init_time + forecast_hour:
            timebase.append((init_time, forecast_hour))

    return timebase


# Helper functions
# =============================================================================


def _to_timedelta(value):
    if isinstance(value, timedelta):
        return value
    else:
        return timedelta(hours=int(value))



"""
Implementation of the "MEM2" method:

see Kim1995:

    Kim, T., Lin, L. H., & Wang, H. (1995). Application of maximum entropy method
    to the real sea data. In Coastal Engineering 1994 (pp. 340-355).

    link: https://icce-ojs-tamu.tdl.org/icce/index.php/icce/article/download/4967/4647
    (working as of May 29, 2022)

and references therein.

"""
import numpy
from scipy.optimize import root
import typing
from numba import njit
from .utils import get_direction_increment


@njit(cache=True)
def nonlinear_equations_for_lagrange_multipliers(
        lambdas, sine_and_cosine, a1, b1, a2, b2,direction_increment):
    """
    Construct the nonlinear equations we need to solve for lambda.

    :param lambdas:
    :param sine_and_cosine:
    :param a1:
    :param b1:
    :param a2:
    :param b2:
    :param direction_increment:
    :return:
    """
    vect = numpy.zeros(sine_and_cosine.shape[1])
    for jj in range(0, 4):
        vect = vect + lambdas[jj] * sine_and_cosine[jj, :]

    dist = numpy.exp(- vect)
    out = numpy.zeros(4)
    out[0] = numpy.sum(
        (a1 - sine_and_cosine[0, :]) * dist * direction_increment
    )
    out[1] = numpy.sum(
        (b1 - sine_and_cosine[1, :]) * dist * direction_increment
    )
    out[2] = numpy.sum(
        (a2 - sine_and_cosine[2, :]) * dist * direction_increment
    )
    out[3] = numpy.sum(
        (b2 - sine_and_cosine[3, :]) * dist * direction_increment)
    return out

@njit(cache=True)
def reconstruction(
        lambdas,
        direction_increment,
        sine_and_cosine
):

    """
    Given the solution for the Lagrange multipliers- reconstruct the directional
    distribution.
    :param lambdas:
    :param direction_increment:
    :param sine_and_cosine:
    :return:
    """
    sum_vector = numpy.zeros(sine_and_cosine.shape[1])
    for jj in range(0, 4):
        sum_vector = sum_vector + lambdas[jj] * sine_and_cosine[jj, :]

    lambda0 = numpy.log(
        numpy.sum(numpy.exp(-sum_vector) * direction_increment)
    )

    return numpy.exp(
        - (sum_vector + lambda0)
    )

@njit(cache=True)
def initial_value(a1, b1, a2, b2):
    """
    Initial guess of the Lagrange Multipliers according to the "MEM AP2" approximation
    found im Kim1995

    :param a1:
    :param b1:
    :param a2:
    :param b2:
    :return:
    """
    guess = numpy.empty((len(a1), 4))
    fac = 1 + a1 ** 2 + b1 ** 2 + a2 ** 2 + b2 ** 2
    guess[:, 0] = (2 * a1 * a2 + 2 * b1 * b2 - 2 * a1 * fac)
    guess[:, 1] = (2 * a1 * b2 - 2 * b1 * a2 - 2 * b1 * fac)
    guess[:, 2] = (a1 ** 2 - b1 ** 2 - 2 * a2 * fac)
    guess[:, 3] = (2 * a1 * b1 - 2 * b2 * fac)
    return guess


def mem2(directions_radians: numpy.ndarray,
         a1: typing.Union[numpy.ndarray, float],
         b1: typing.Union[numpy.ndarray, float],
         a2: typing.Union[numpy.ndarray, float],
         b2: typing.Union[
             numpy.ndarray, float]) -> numpy.ndarray:
    """
    Return the directional distribution that maximizes Shannon [ - D log(D) ]
    enthrophy constrained by given observed directional moments,

    :param directions_radians: 1d array of wave directions in radians,
    length[number_of_directions]

    :param a1: 1d array of cosine directional moment as function of frequency,
    length [number_of_frequencies]

    :param b1: 1d array of sine directional moment as function of frequency,
    length [number_of_frequencies]

    :param a2: 1d array of double angle cosine directional moment as function
    of frequency, length [number_of_frequencies]

    :param b2: 1d array of double angle sine directional moment as function of
    frequency, length [number_of_frequencies]

    :return: array with shape [number_of_frequencies,number_of_direction]
    representing the directional distribution of the waves at each frequency.

    Maximize the enthrophy of the solution with entrophy defined as:

           integrate - D * log(D) over directions

    such that the resulting distribution D reproduces the observed moments.

    """

    a1 = numpy.atleast_1d(a1)
    b1 = numpy.atleast_1d(b1)
    a2 = numpy.atleast_1d(a2)
    b2 = numpy.atleast_1d(b2)

    number_of_frequencies = len(a1)
    directional_distribution = numpy.zeros(
        (number_of_frequencies, len(directions_radians)))

    direction_increment = get_direction_increment(directions_radians)

    sine_and_cosine = numpy.empty((4, len(directions_radians)))
    sine_and_cosine[0, :] = numpy.cos(directions_radians)
    sine_and_cosine[1, :] = numpy.sin(directions_radians)
    sine_and_cosine[2, :] = numpy.cos(2 * directions_radians)
    sine_and_cosine[3, :] = numpy.sin(2 * directions_radians)

    guess = initial_value(a1, b1, a2, b2)
    for ifreq in range(0, number_of_frequencies):
        #
        res = root(
            nonlinear_equations_for_lagrange_multipliers,
            guess[ifreq, :],
            args=(
                sine_and_cosine, a1[ifreq], b1[ifreq], a2[ifreq], b2[ifreq],
                direction_increment)
        )
        lambas = res.x
        directional_distribution[ifreq, :] = reconstruction(
            lambas, direction_increment, sine_and_cosine)
    return numpy.squeeze(directional_distribution)


import numpy
from pandas import DataFrame
from roguewave.interpolate.dataarray import interpolate_track_data_arrray
from roguewave.interpolate.dataframe import interpolate_dataframe_time
from roguewave.interpolate.geometry import Geometry, convert_to_track_set
from roguewave.interpolate.nd_interp import NdInterpolator
from typing import Dict, Tuple, Mapping
from xarray import Dataset, DataArray

def interpolate_dataset_grid(
        coordinates,
        data_set: Dataset,
        periodic_data: Mapping[str, Tuple[int, int]] = None,
        longitude_variable_in_dataset: str = 'longitude',
) -> Dataset:
    if periodic_data is None:
        periodic_data = {}
        if longitude_variable_in_dataset in data_set:
            periodic_data[longitude_variable_in_dataset] = (360, 180)

        for variable in data_set:
            if 'direction' in str(variable).lower():
                periodic_data[variable] = (360, 360)

    first = True
    return_data_set = None
    for coordinate_name, coordinate_value in coordinates.items():
        if first:
            _data_set = data_set
            first = False
        else:
            _data_set = return_data_set
        return_data_set = interpolate_dataset_along_axis(
            coordinate_value, _data_set,coordinate_name
        )
    return return_data_set


def interpolate_dataset_along_axis(
        coordinate_value,
        data_set: Dataset,
        coordinate_name: str = 'time',
        periodic_data: Mapping[str, Tuple[int, int]] = None,
        periodic_coordinates: Dict = None,
) -> Dataset:
    if periodic_data is None:
        periodic_data = {'longitude':(360,180)}
        for variable in data_set:
            if 'direction' in str(variable).lower():
                periodic_data[variable] = (360, 360)

    if periodic_coordinates is None:
        periodic_coordinates = {'longitude':360,'direction':360}

    points = {
        coordinate_name: coordinate_value
    }
    return_data_set = Dataset()

    for variable in data_set:
        if coordinate_name not in list(data_set[variable].coords):
            return_data_set[variable] = data_set[variable]
            continue

        coordinate_names = [str(x) for x in data_set[variable].dims]
        data_coordinates = [ (name,data_set[variable].coords[name].values)
                         for name in coordinate_names ]

        period_data, discont_data = None, None
        if variable in periodic_data:
            period_data, discont_data = periodic_data[variable]

        dimensions = list(data_set[variable].dims)

        def get_data(indices,idims):
            index = [slice(None)] * len(dimensions)
            for interp_index,idim in zip(indices,idims):
                index[idim] = interp_index
            return data_set[variable].values[tuple(index)]

        data_accessing = NdInterpolator(get_data,
                                        data_coordinates,
                                        data_set[variable].shape,
                                        list(points.keys()),
                                        coordinate_name,
                                        periodic_coordinates,
                                        period_data,
                                        discont_data
                                        )

        coor = { value[0]:value[1] for value in data_coordinates }
        for coor_name,coor_value in points.items():
            coor[coor_name] = coor_value

        return_data_set[variable] = DataArray(
            data=data_accessing.interpolate(points),
            coords=coor,
            dims=data_set[variable].dims
        )

    return return_data_set


def interpolate_dataset(
        data_set: Dataset,
        geometry: Geometry,
        periodic_coordinates: Dict[str, float] = None,
        periodic_data: Dict[str, Tuple[float, float]] = None,
        time_variable_in_dataset: str = 'time',
        longitude_variable_in_dataset: str = 'longitude',
        latitude_variable_in_dataset: str = 'latitude',
):
    geometry = convert_to_track_set(
        geometry, data_set[time_variable_in_dataset].values)

    if periodic_coordinates is None:
        periodic_coordinates = {longitude_variable_in_dataset: 360}

    for variable in data_set:
        if 'direction' in str(variable).lower():
            periodic_data = {variable: (360, 360)}

    out = {}
    for name, track in geometry.tracks.items():
        points = {
            time_variable_in_dataset: track.time,
            longitude_variable_in_dataset: track.longitude,
            latitude_variable_in_dataset: track.latitude
        }
        out[name] = interpolate_at_points(
            data_set=data_set,
            points=points,
            independent_variable=time_variable_in_dataset,
            periodic_coordinates=periodic_coordinates,
            periodic_data=periodic_data
        ).to_dataframe()
        if out[name].index.tz is None:
            out[name].index = out[name].index.tz_localize('utc')
    return out


def interpolate_at_points(
        data_set: Dataset,
        points: Dict[str, numpy.ndarray],
        independent_variable=None,
        periodic_coordinates: Dict[str, float] = None,
        periodic_data: Dict[str, Tuple[float, float]] = None
) -> Dataset:
    if periodic_data is None:
        periodic_data = {}

    dimensions = data_set.dims
    if independent_variable is None:
        if 'time' in dimensions:
            independent_variable = 'time'
        else:
            independent_variable = dimensions[0]

    return_data_set = Dataset(
        coords={independent_variable: points[independent_variable]}
    )
    for variable in data_set:
        if variable in periodic_data:
            period_data = periodic_data[variable][0]
            discont = periodic_data[variable][1]
        else:
            period_data = None
            discont = None

        return_data_set[variable] = interpolate_track_data_arrray(
            data_set[variable], points, independent_variable,
            periodic_coordinates=periodic_coordinates,
            period_data=period_data,
            discont=discont
        )
    return return_data_set


def tracks_as_dataset(time, drifter_tracks: Mapping[str, DataFrame])->DataArray:
    tracks = {}
    for track_id, track in drifter_tracks.items():
        variables = list(track.columns)

        tracks[track_id] = DataArray(
            interpolate_dataframe_time(track, time),
            coords={'time': time, 'variables': variables},
            dims=['time', 'variables'],
            name=track_id,
        )

    dataset = Dataset(tracks).to_array(dim='spotter_id')
    return dataset



"""
Contents: Routines to get data from the spotter api

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Routines to get data from the spotter api

Functions:

- `get_spectrum`, function to download spectral data.
- `get_bulk_wave_data`, function to download bulk wave data.
- `get_data`, general function to download different data types (spectral,
    bulk wave, wind, SST, barometer).
- `search_circle`, get all available data within a given circle.
- `search_rectangle`, get all available data within a given rectangle.

"""

# 1) Imports
# =============================================================================
from datetime import datetime, timedelta, timezone
from .exceptions import ExceptionNoDataForVariable
from multiprocessing.pool import ThreadPool
from pysofar.spotter import Spotter, SofarApi
from roguewave import logger
from roguewave.tools.time import datetime_to_iso_time_string, to_datetime_utc
from roguewave.wavespectra  import FrequencySpectrum, create_1d_spectrum

from roguewave.metoceandata import WaveBulkData, as_dataframe, WindData, \
    SSTData, MetoceanData, BarometricPressure
from typing import Dict, List, Union, TypedDict, Tuple, Sequence
from tqdm import tqdm
from pandas import DataFrame
from .helper_functions import _get_sofar_api, get_spotter_ids
import numpy
import roguewave.spotterapi.spotter_cache as spotter_cache
from xarray import Dataset, concat

# 2) Constants & Private Variables
# =============================================================================
# Maximum number of spectra to retrieve from the Spotter API per API call. Note
# that 2- os a hard limit of the API. If set higher than 100 it will just
# return 100 (and the implementation will fail)
MAX_LOCAL_LIMIT = 100
MAX_LOCAL_LIMIT_BULK = 500

# Maximum number of workers in the Threadpool. Should be set to something
# reasonable to not overload wavefleet
MAXIMUM_NUMBER_OF_WORKERS = 40

# Number of retry attemps if a download fails.
NUMBER_OF_RETRIES = 2


# 3) Classes
# =============================================================================
class VariablesToInclude(TypedDict):
    """
    Dictionary that is used to indicate which data to download. The keys
    correspond exactly to the categories returned by a wavefleet request.
    As a consequence we also use the keys as an enumeration of which variables
    are potentially available. This class is internal- should not be
    instantiated or returned.
    """
    frequencyData: bool
    waves: bool
    wind: bool
    surfaceTemp: bool
    barometerData: bool


# 4) Main Functions
# =============================================================================
def get_spectrum(
        spotter_ids: Union[str, Sequence[str]],
        start_date: Union[datetime, int, float, str] = None,
        end_date: Union[datetime, int, float, str] = None,
        session: SofarApi = None,
        parallel_download=True,
        cache: bool = True
) -> Dict[str, FrequencySpectrum]:
    """
    Gets the requested frequency wave data for the spotter(s) in the given
    interval

    :param spotter_ids: Can be either 1) a List of spotter_ids or 2) a single
    Spotter_id.

    :param start_date: ISO 8601 formatted date string, epoch or datetime.
                       If not included defaults to beginning of spotters
                       history

    :param end_date:   ISO 8601 formatted date string, epoch or datetime.
                       If not included defaults to end of spotter history

    :param session:    Active SofarApi session. If none is provided one will be
                       creatated automatically. This requires that an API key
                       is set in the environment.

    :param parallel_download: Use multiple requests to the Api to speed up
                        retrieving data. Only useful for large requests.

    :param cache: Cache requests. If True, returned data will be stored in
                        a file Cache on disk, and repeated calls with the
                        same arguments will use locally cached data. The cache
                        is a FileCache with a maximum of 2GB by default.

    :return: Data as a dictornary with spotter_id's as keys, and for each
    corresponding value a List that for each returned time contains a
    WaveSpectrum1D object.

    """
    data = get_data(
        spotter_ids,
        start_date,
        end_date,
        include_frequency_data=True,
        include_waves=False,
        include_wind=False,
        include_surface_temp_data=False,
        session=session,
        parallel_download=parallel_download,
        cache=cache
    )
    out = {}
    for key in data:
        out[key] = data[key]['frequencyData']
    return out


# -----------------------------------------------------------------------------


def get_bulk_wave_data(
        spotter_ids: Union[str, Sequence[str]],
        start_date: Union[datetime, int, float, str] = None,
        end_date: Union[datetime, int, float, str] = None,
        session: SofarApi = None,
        parallel_download: bool = True,
        bulk_data_as_dataframe: bool = True,
        cache: bool = True,
        convert_to_sofar_model_names = False
) -> Dict[str, Union[DataFrame,list[WaveBulkData]]]:
    """
    Gets the requested bulk wave data for the spotter(s) in the given interval

    :param spotter_ids: Can be either 1) a List of spotter_ids or 2) a single
    Spotter_id.

    :param start_date: ISO 8601 formatted date string, epoch or datetime.
                       If not included defaults to beginning of spotters
                       history

    :param end_date:   ISO 8601 formatted date string, epoch or datetime.
                       If not included defaults to end of spotter history

    :param session:    Active SofarApi session. If none is provided one will be
                       creatated automatically. This requires that an API key
                       is set in the environment.

    :param parallel_download: Use multiple requests to the Api to speed up
                        retrieving data. Only useful for large requests.

    :param bulk_data_as_dataframe: return bulk data as a dataframe per Spotter
                       instead of a list of BulkWaveVariable objects.
                       Defaults to yes- only set to false for dev purposes.

    :param cache: Cache requests. If True, returned data will be stored in
                        a file Cache on disk, and repeated calls with the
                        same arguments will use locally cached data. The cache
                        is a FileCache with a maximum of 2GB by default.

    :return: Data as a dictornary with spotter_id's as keys, and for each
    corresponding value a dataframe containing the output.
    """
    data = get_data(
        spotter_ids,
        start_date,
        end_date,
        include_frequency_data=False,
        include_waves=True,
        include_wind=False,
        include_surface_temp_data=False,
        session=session,
        parallel_download=parallel_download,
        bulk_data_as_dataframe=bulk_data_as_dataframe,
        cache=cache,
        convert_to_sofar_model_names=convert_to_sofar_model_names
    )
    out = {}
    for key in data:
        out[key] = data[key]['waves']
    return out


# -----------------------------------------------------------------------------
def get_data(
        spotter_ids: Union[str, List[str]],
        start_date: Union[datetime, int, float, str] = None,
        end_date: Union[datetime, int, float, str] = None,
        include_frequency_data=False,
        include_waves=True,
        include_wind=False,
        include_barometer_data=False,
        include_surface_temp_data=False,
        session: SofarApi = None,
        parallel_download=True,
        bulk_data_as_dataframe=True,
        spectral_data_as_dataset_spectrum=True,
        cache=True,
        convert_to_sofar_model_names=False
) -> Dict[str, Dict[str, Union[FrequencySpectrum,
                               list[WaveBulkData], DataFrame]]]:
    """
    Gets the requested data for the spotter(s) in the given interval

    :param spotter_ids: Can be either 1) a List of spotter_ids or 2) a single
    Spotter_id.

    :param start_date: ISO 8601 formatted date string, epoch or datetime.
                       If not included defaults to beginning of spotters
                       history

    :param end_date:   ISO 8601 formatted date string, epoch or datetime.
                       If not included defaults to end of spotter history

    :param include_frequency_data: set to True to return spectral data
        (if available)

    :param include_waves: set to True to return wave data (if available)

    :param include_wind: set to True to return wind data (if available)

    :param include_surface_temp_data: set to True to return SST data
        (if available)

    :param include_barometer_data: set to True to return barometer data
        (if available)

    :param session:    Active SofarApi session. If none is provided one will be
                       created automatically. This requires that an API key is
                       set in the environment.

    :param parallel_download: Use multiple requests to the Api to speed up
                       retrieving data. Only useful for large requests.

    :param bulk_data_as_dataframe: return bulk data as a dataframe per Spotter
                       instead of a list of BulkWaveVariable objects. Defaults
                       to yes- only set to false for dev purposes.

    :param cache: Cache requests. If True, returned data will be stored in
                        a file Cache on disk, and repeated calls with the
                        same arguments will use locally cached data. The cache
                        is a FileCache with a maximum of 2GB by default.

    :return: Data as a dictornary with spotter_id's as keys, and for each
    corresponding value a dataframe containing the output.

    To Note; this function now fetches each of the variables seperately -
        instead of doing one fetch; this is a clutch fix to the issue that
        data for different variables have different timestamps. As a consequence
        getting 100 data points for frequency data may cover 5 days -whereas
        the same number of points only cover a day for SST data. Since we
        only get 100 data points per request- and have to advance to the next
        date ourselves-  this complicates matters tremendously.
    """

    variables_to_include = VariablesToInclude(
        waves=include_waves,
        wind=include_wind,
        surfaceTemp=include_surface_temp_data,
        frequencyData=include_frequency_data,
        barometerData=include_barometer_data,
    )
    out = {}
    for variable,to_include in variables_to_include.items():
        if to_include:
            print(
                f"Getting spotter data: retrieving {variable}")
            dummy = VariablesToInclude(
                frequencyData=False,
                wind=False,
                waves=False,
                surfaceTemp=False,
                barometerData=False)
            dummy[variable] = to_include

            temp = _get_data(
                spotter_ids,
                start_date,
                end_date,
                dummy['frequencyData'],
                dummy['waves'],
                dummy['wind'],
                dummy['barometerData'],
                dummy['surfaceTemp'],
                session,
                parallel_download,
                bulk_data_as_dataframe,
                spectral_data_as_dataset_spectrum,
                cache,
                convert_to_sofar_model_names
            )
            for spotter_id in temp:
                if spotter_id not in out:
                    out[spotter_id] = {}
                out[spotter_id][variable] = temp[spotter_id][variable]
    return out

def _get_data(
        spotter_ids: Union[str, List[str]],
        start_date: Union[datetime, int, float, str] = None,
        end_date: Union[datetime, int, float, str] = None,
        include_frequency_data=False,
        include_waves=True,
        include_wind=False,
        include_barometer_data=False,
        include_surface_temp_data=False,
        session: SofarApi = None,
        parallel_download=True,
        bulk_data_as_dataframe=True,
        spectral_data_as_dataset_spectrum=True,
        cache=True,
        convert_to_sofar_model_names=False
) -> Dict[str, Dict[str, Union[FrequencySpectrum,
                               list[WaveBulkData], DataFrame]]]:
    """
    Gets the requested data for the spotter(s) in the given interval

    :param spotter_ids: Can be either 1) a List of spotter_ids or 2) a single
    Spotter_id.

    :param start_date: ISO 8601 formatted date string, epoch or datetime.
                       If not included defaults to beginning of spotters
                       history

    :param end_date:   ISO 8601 formatted date string, epoch or datetime.
                       If not included defaults to end of spotter history

    :param include_frequency_data: set to True to return spectral data
        (if available)

    :param include_waves: set to True to return wave data (if available)

    :param include_wind: set to True to return wind data (if available)

    :param include_surface_temp_data: set to True to return SST data
        (if available)

    :param include_barometer_data: set to True to return barometer data
        (if available)

    :param session:    Active SofarApi session. If none is provided one will be
                       created automatically. This requires that an API key is
                       set in the environment.

    :param parallel_download: Use multiple requests to the Api to speed up
                       retrieving data. Only useful for large requests.

    :param bulk_data_as_dataframe: return bulk data as a dataframe per Spotter
                       instead of a list of BulkWaveVariable objects. Defaults
                       to yes- only set to false for dev purposes.

    :param cache: Cache requests. If True, returned data will be stored in
                        a file Cache on disk, and repeated calls with the
                        same arguments will use locally cached data. The cache
                        is a FileCache with a maximum of 2GB by default.

    :return: Data as a dictornary with spotter_id's as keys, and for each
    corresponding value a dataframe containing the output.
    """

    if spotter_ids is None:
        spotter_ids = get_spotter_ids()

    variables_to_include = VariablesToInclude(
        waves=include_waves,
        wind=include_wind,
        surfaceTemp=include_surface_temp_data,
        frequencyData=include_frequency_data,
        barometerData=include_barometer_data,
    )

    # Make sure we have a list object
    if not isinstance(spotter_ids, list):
        spotter_ids = [spotter_ids]

    if cache and end_date < datetime.now(tz=timezone.utc):
        # Use the cache _only_ for requests that concern the past. In real-time
        # things may change.

        out = spotter_cache.get_data(
            spotter_ids,
            session,
            _download_data,
            variables_to_include=variables_to_include,
            start_date=start_date,
            end_date=end_date,
            bulk_data_as_dataframe=bulk_data_as_dataframe,
            convert_to_sofar_model_names=convert_to_sofar_model_names)
    else:
        if session is None:
            session = _get_sofar_api()

        def worker(_spotter_id):
            return _download_data(_spotter_id, session, variables_to_include,
                                  start_date, end_date, bulk_data_as_dataframe,
                                  limit=None,
                                  convert_to_sofar_model_names=
                                  convert_to_sofar_model_names)

        if parallel_download:
            with ThreadPool(processes=MAXIMUM_NUMBER_OF_WORKERS) as pool:
                out = list(
                    tqdm(pool.imap(worker, spotter_ids),
                         total=len(spotter_ids)))
        else:
            out = list(
                tqdm(map(worker, spotter_ids), total=len(spotter_ids)))

    data = {}
    for spotter_id, spotter_data in zip(spotter_ids, out):
        #
        # Did we get any data for this spotter
        if spotter_data is not None:
            if 'frequencyData' in spotter_data:
                if isinstance(spotter_data['frequencyData'],List):
                    # This will be deprecated in a future version.
                    spotter_data['frequencyData'] = \
                        FrequencySpectrum.concat_from_list(
                            spotter_data['frequencyData'])
            data[spotter_id] = spotter_data

    return data


# -----------------------------------------------------------------------------


def search_circle(
        start_date: Union[datetime, str],
        end_date: Union[datetime, str],
        center_lat_lon: Tuple,
        radius: float,
        session: SofarApi = None,
        cache=False
):
    """
    Search for all Spotters that have data available within the give spatio-
    temporal region defined by a circle with given center and radius and start-
    and end- dates. This calls the "search" endpoint of wavefleet.

    :param start_date: ISO 8601 formatted date string, epoch or datetime.
    :param end_date:   ISO 8601 formatted date string, epoch or datetime.
    :param center_lat_lon: (Latitude, Longitude) of the center of the circle.
    :param radius: Radius in meter of the circle.
    :param session: Active SofarApi session. If none is provided one will be
                    created automatically. This requires that an API key is
                    set in the environment.
    :return:
    """

    geometry = {'type': 'circle', 'points': center_lat_lon, 'radius': radius}

    if session is None:
        session = _get_sofar_api()

    print("Get Spotter data: retrieving all data from spatio-temporal region")
    if cache:
        return spotter_cache.get_data_search(
            handler=_search,
            session=session,
            start_date=start_date,
            end_date=end_date,
            geometry=geometry
        )
    else:
        #
        return _search(start_date, end_date, geometry, session)


# -----------------------------------------------------------------------------


def search_rectangle(
        start_date: Union[datetime, str],
        end_date: Union[datetime, str],
        bounding_box,
        session: SofarApi = None,
        cache=False
):
    """
    Search for all Spotters that have data available within the give spatio-
    temporal region defined by a circle with given center and radius and start-
    and end-dates. This calls the "search" endpoint of wavefleet.

    :param start_date: ISO 8601 formatted date string, epoch or datetime.
    :param end_date:   ISO 8601 formatted date string, epoch or datetime.
    :param bounding_box: coordinates of two points that define a rectangular
    bounding box. Coordinates per point are given as (lat, lon) pairs, and the
    input takes the form of a list/tuple of points: ( (p1_lat, p1_lon),(p2_lat,
    p2_lon) )
    :param session: Active SofarApi session. If none is provided one will be
                    created automatically. This requires that an API key is
                    set in the environment.
    :return:
    """
    geometry = {'type': 'envelope',
                'points': bounding_box,
                'radius': None}

    if session is None:
        session = _get_sofar_api()

    print("Get Spotter data: retrieving all data from spatio-temporal region")
    if cache:
        return spotter_cache.get_data_search(
            _search,
            session,
            start_date=start_date,
            end_date=end_date,
            geometry=geometry
        )
    else:
        #
        return _search(start_date, end_date, geometry, session)


# -----------------------------------------------------------------------------


# 5) Private Functions
# =============================================================================
def _download_data(
        spotter_id: str,
        session: SofarApi,
        variables_to_include: VariablesToInclude,
        start_date: Union[datetime, str, int, float] = None,
        end_date: Union[datetime, str, int, float] = None,
        bulk_data_as_dataframe: bool = True,
        limit: int = None,
        convert_to_sofar_model_names=False) -> \
        Dict[str, Union[List[FrequencySpectrum], List[WaveBulkData]]]:
    """
    Function that downloads data from the API for the requested Spotter
    It abstracts away the limitation that the API can only return a maximum
    of 100 Spectra or 500 bulk data points for a single Spotter per call.

    :param spotter_id: ID for the Spotter we want to download
    :param session: Active session of the SofarAPI
    :param start_date: ISO 8601 formatted date string, epoch or datetime.
                       If not included defaults to beginning of spotters
                       history
    :param end_date: ISO 8601 formatted date string, epoch or datetime.
                     If not included defaults to end of spotter history
    :param limit: Maximum number of Spectra to download per call. Not exposed
                  externally right now as there is not really a reason to
                  change it. (calling function does not set the limit)

    :return: List of available data in the requested timeframe.
    """

    # Create a Spotter object to Query
    spotter = Spotter(spotter_id, spotter_id, session=session)

    # Set the initial start date. This will get advanced for every 100 spectra
    # we download
    _start_date = start_date

    if variables_to_include['frequencyData']:
        max_local_limit = MAX_LOCAL_LIMIT
    else:
        max_local_limit = MAX_LOCAL_LIMIT_BULK

    data = {}
    while True:
        # We can only download a maximum of 100 spectra at a time; so we need
        # to loop our request. We do not know how many spotters there are
        # in the given timeframe.
        #
        # Assumptions:
        #   - spotter api returns a maximum of N items per requests (N=500 if
        #     no spectral data, N=100 if so)
        #   - requests returned start from the requested start data and
        #     with the last entry either being the last entry that fits
        #     in the requested window, or merely the last sample that fits
        #     in the 100 items.
        #   - requests returned are in order

        if limit is not None:
            # If we hit the limit (if given) of spotters requested, break
            if len(data) >= limit:
                break

            # otherwise, update our next request so we don't overrun
            # the limit
            local_limit = min(limit - len(data), max_local_limit)
        else:
            # if no limit is given, just ask for the maximum allowed.
            local_limit = max_local_limit

        try:
            # Try to get the next batch of data
            _next, number_of_items_returned, max_time = \
                _get_next_page(spotter, variables_to_include,
                               _start_date, end_date, local_limit)

        except ExceptionNoDataForVariable as e:
            # Could not download data, add nothing, raise warning
            raise e

        # Add the data to the list

        for key in _next:
            if key in data:
                data[key] += _next[key]
            else:
                data[key] = _next[key]

        # If we did not receive all data we requested...
        if number_of_items_returned < local_limit:
            # , we are done...
            break
        else:
            # ... else we update the startdate to be the time of the last
            # known entry we received plus a second, and use this as the new
            # start.
            _start_date = to_datetime_utc(max_time,to_scalar=True) + timedelta(seconds=1)

    # Postprocessing
    if len(data) < 1:
        data = None
    else:
        if bulk_data_as_dataframe:
            # Convert bulk data to a dataframe if desired
            for key in data:
                if key == 'frequencyData':
                    # Concatenate all the spectral dataset along time dimension
                    # and return a wave frequency spectrum object
                    data[key] = FrequencySpectrum.concat_from_list(data[key])
                else:
                    data[key] = as_dataframe(
                        data[key],convert_to_sofar_model_names)
    return data


# -----------------------------------------------------------------------------


def _get_next_page(
        spotter: Spotter,
        variables_to_include: VariablesToInclude,
        start_date: Union[datetime, str, int, float] = None,
        end_date: Union[datetime, str, int, float] = None,
        limit: int = MAX_LOCAL_LIMIT,
) -> Tuple[Dict[str, Union[List[Dataset],
                           List[WaveBulkData]]], int, datetime]:
    """
    Function that downloads the page of Data from the Spotter API that lie
    within the given interval, starting from the record closest to the
    startdate.

    idiosyncrasies to handle:
    - Wavefleet sometimes returns multiple instances of the same record (same
      time). These are filtered through a all to _unique. This should
      be fixed in the future.
    - Some entries are broken (None for entries).
    - Not all Spotters will have (all) data for the given timerange.
    - Wavefleet sometimes times out on requests.

    :param spotter: Spotter object from pysofar

    :param start_date: ISO 8601 formatted date string, epoch or datetime.
                       If not included defaults to beginning of spotters
                       history

    :param end_date: ISO 8601 formatted date string, epoch or datetime.
                     If not included defaults to end of spotter history

    :param limit: Maximum number of Spectra to download per call. Not exposed
                  externally right now as there is not really a reason to
                  change it. (calling function does not set the limit)

    :return: Data
    """

    # Retry mechanism
    start_date = to_datetime_utc(start_date)
    end_date = to_datetime_utc(end_date)
    json_data = None
    for retry in range(0, NUMBER_OF_RETRIES + 1):
        try:
            json_data = spotter.grab_data(
                limit=limit,
                start_date=datetime_to_iso_time_string(start_date),
                end_date=datetime_to_iso_time_string(end_date),
                include_frequency_data=variables_to_include[
                    'frequencyData'],
                include_directional_moments=variables_to_include[
                    'frequencyData'],
                include_barometer_data=variables_to_include['barometerData'],
                include_waves=variables_to_include['waves'],
                include_wind=variables_to_include['wind'],
                include_surface_temp_data=variables_to_include[
                    'surfaceTemp'],
            )
            break
        except Exception as e:
            if retry < NUMBER_OF_RETRIES:
                warning = f'Error downloading data for {spotter.id}, ' \
                          f'attempting retry {retry + 1}'
                logger.warning(warning)
            else:
                raise e

    out = {}
    max_num_items = 0
    max_time = datetime(1970, 1, 1, tzinfo=timezone.utc)
    #
    # Loop over all possible variables
    for var_name, include_variable in variables_to_include.items():
        #
        # Did we want to include this variable?
        if include_variable:
            #
            # If so- was it returned? If not- raise error
            if var_name not in json_data:
                raise ExceptionNoDataForVariable(var_name)

            # If so- were any elements returned for this period? If not
            # continue. We could error here, but there may be gaps in data for
            # certain sensors, so we try to continue
            if not json_data[var_name]:
                continue

            elements_returned = len(json_data[var_name])

            # Filter for Nones.
            json_data[var_name] = _none_filter(json_data[var_name])
            if not json_data[var_name]:
                # If after filtering for bad data nothing remains, continue.
                continue

            # How many items were returned (counting doubles and none values,
            # this is used for the pagination logic which only knows if it is
            # done if less then the requested data was returned).
            max_num_items = max(max_num_items, elements_returned)

            # Add to output
            out[var_name] = \
                [_get_class(var_name, data) for data in json_data[var_name]]

            # Filter for doubles.
            out[var_name] = _unique_filter(out[var_name])

            # get the new max time. Our object can be a FrequencySpectrum or
            # a metocean data object here. The latter returns a scalar date-
            # time object, the former returns a length 1 data-array of
            # datetime64. To ensure we get a datetime - we call the datetime
            # conversion and request output as a scalar.
            if to_datetime_utc(out[var_name][-1].time,True) > max_time:
                max_time = to_datetime_utc(out[var_name][-1].time, True)

    return out, max_num_items, max_time


# -----------------------------------------------------------------------------


def _search(start_date: Union[datetime, str],
            end_date: Union[datetime, str],
            geometry: dict,
            session: SofarApi = None,
            variables_to_include: VariablesToInclude = None,
            page_size=500,
            bulk_data_as_dataframe: bool = True,
            ):
    if session is None:
        session = _get_sofar_api()

    if variables_to_include is None:
        variables_to_include = VariablesToInclude(
            frequencyData=True, waves=True, wind=True, surfaceTemp=True,
            barometerData=True
        )

    start_date_str = datetime_to_iso_time_string(start_date)
    end_date_str = datetime_to_iso_time_string(end_date)

    shape = geometry['type']
    shape_params = geometry['points']
    radius = geometry['radius']

    generator = session.search(
        shape=shape,
        shape_params=shape_params,
        start_date=start_date_str,
        end_date=end_date_str,
        radius=radius,
        page_size=page_size,
        return_generator=True
    )

    spotters = {}
    # loop over all spotters returned
    for spotter in generator:
        spotter_id = spotter['spotterId']
        # loop over keys we can parse
        for key, value in variables_to_include.items():
            #
            if not value:
                # Do we want to include this variable?
                continue

            if key in spotter:
                item = spotter[key]
                if not item:
                    # no data
                    continue

                item['latitude'] = spotter['latitude']
                item['longitude'] = spotter['longitude']
                item['timestamp'] = spotter['timestamp']
                data = _get_class(key, item)
                if spotter_id not in spotters:
                    spotters[spotter_id] = {}

                if key not in spotters[spotter_id]:
                    spotters[spotter_id][key] = []

                spotters[spotter_id][key].append(data)

    for spotter_id in spotters:
        for key in spotters[spotter_id]:
            # ensure results are unique
            spotters[spotter_id][key] = _unique_filter(
                spotters[spotter_id][key])

            if bulk_data_as_dataframe and (not key == 'frequencyData'):
                spotters[spotter_id][key] = as_dataframe(
                    spotters[spotter_id][key])

            if key == 'frequencyData':
                spotters[spotter_id][key] = FrequencySpectrum.concat_from_list(
                    spotters[spotter_id][key]
                )
    return spotters


# -----------------------------------------------------------------------------


# 6) Helper Functions
# =============================================================================
def _unique_filter(data):
    """
    Filter for dual time entries that occur due to bugs in wavefleet (same
    record returned twice)
    :param data:
    :return:
    """

    # Get time
    if isinstance(data[0],FrequencySpectrum ):
        time = numpy.array(
            [to_datetime_utc(x['time'].values[0]).timestamp() for x in data])
    else:
        time= numpy.array([x.time.timestamp() for x in data])

    # Get indices of unique times
    _, unique_indices = numpy.unique(time, return_index=True)

    # Return only unique indices
    return [data[index] for index in unique_indices]


# -----------------------------------------------------------------------------


def _none_filter(data: Dict):
    """
    Filter for the occasional occurance of bad data returned from wavefleet.
    :param data:
    :return:
    """
    return list(
        filter(
            lambda x: (x['latitude'] is not None) and (
                    x['longitude'] is not None) and (
                                  x['timestamp'] is not None),
            data
        )
    )


# -----------------------------------------------------------------------------


def _get_class(key, data) -> Union[MetoceanData, FrequencySpectrum]:
    if key == 'waves':
        return WaveBulkData(
            time=data['timestamp'],
            latitude=data['latitude'],
            longitude=data['longitude'],
            significant_waveheight=data['significantWaveHeight'],
            peak_period=data['peakPeriod'],
            mean_period=data['meanPeriod'],
            peak_direction=data['peakDirection'],
            peak_directional_spread=data['peakDirectionalSpread'],
            mean_direction=data['meanDirection'],
            mean_directional_spread=data['meanDirectionalSpread'],
            peak_frequency=1.0 / data['peakPeriod']
        )
    elif key == 'frequencyData':
        return create_1d_spectrum(
            frequency=numpy.array(data['frequency']),
            variance_density=numpy.array(data['varianceDensity']),
            time=to_datetime_utc(data['timestamp']),
            latitude=numpy.array(data['latitude']),
            longitude=numpy.array(data['longitude']),
            a1=numpy.array(data['a1']),
            b1=numpy.array(data['b1']),
            a2=numpy.array(data['a2']),
            b2=numpy.array(data['b2'])
            )
    elif key == 'wind':
        data = data.copy()
        data['time'] = data['timestamp']
        data.pop('timestamp')
        return WindData(**data)
    elif key == 'surfaceTemp':
        data = data.copy()
        data['time'] = data['timestamp']
        data.pop('timestamp')
        return SSTData(**data)
    elif key == 'barometerData':
        data = data.copy()
        data['time'] = data['timestamp']
        data.pop('timestamp')
        return BarometricPressure(**data)
    else:
        raise Exception('Unknown variable')
# -----------------------------------------------------------------------------



"""
This module is primarily intended to read model setup data from a model
definition file generated by WW3 (v6.07). There exists (to my knowledge) no
formal definition of a ww3 model_def file - as the format is primarily intended
to be used by ww3 internally. Here we need the information in the model_def
file so we can read restart files. The current implementation was generated
from reverse engineering from the Fortran source how we should interpret the
binary, unformatted sequential file (Fortran writing type). We do _not_ try
and read all data (some is ignored)- only an effort is made to read what we
need for restart files.

To note:
In case of changes in ww3 that change the format it may be easier to instead
of rewriting an interpreter just providing the needed information directly.
Specifically we need:
- frequencies
- directions
- depth
- relation linear index to ilat, ilon
- relation ilat, ilon to linear index
- number of "sea" or spatial points.

All of these can in principle be inferred form the way the model is set-up. We
do not _need_ to read the model_def file (though it is convinient to do this
automatically).

Pieter Smit, Aug, 2022

"""

import numpy
from roguewave.wavewatch3.fortran_types import FortranInt, FortranFloat, \
    FortranCharacter
from roguewave.wavewatch3.resources import Resource
from io import BytesIO
from dataclasses import dataclass
from typing import Union, Tuple, Sequence, Literal
from roguewave.tools.math import wrapped_difference


@dataclass()
class Grid:
    """
    A class that encapsulates all grid data from the model definition. It also
    contains convinience methods to project linear indexed data to a 2d array
    of latitude and longitude (with NaN for missing values), and vice versa
    extract the needed data to construct a linear indexed version compatible
    with WW3.

    Purpose:
    We need this information for reading restart files.
    """
    number_of_spatial_points: int  # number of spatial "sea" points
    frequencies: numpy.ndarray
    directions: numpy.ndarray
    latitude: numpy.ndarray  # 1D array of model latitudes
    longitude: numpy.ndarray # 1D array of model longitudes
    _to_linear_index: numpy.ndarray  # mapping of [ilon,ilat] => linear index
    _to_point_index: numpy.ndarray  # mapping of linear index => [ilon,ilat]

    @property
    def _growth_factor(self):
        return self.frequencies[1]/self.frequencies[0]

    def longitude_index(self, linear_index:Union[slice,Sequence,numpy.ndarray]
                        )->numpy.ndarray:
        """
        :param linear_index: linear 1d index
        :return: longitude 1d index
        """

        return self._to_point_index[0, linear_index]

    def latitude_index(self, linear_index:Union[slice,Sequence,numpy.ndarray]
                        )->numpy.ndarray:
        """
        :param linear_index: linear 1d index
        :return: latitude 1d index
        """
        return self._to_point_index[1, linear_index]

    def index(self,
              latitude_index:Union[slice,Sequence,numpy.ndarray],
              longitude_index:Union[slice,Sequence,numpy.ndarray],
              valid_only=False):
        """
        get the linear index of the array
        :param latitude_index:
        :param longitude_index:
        :return:
        """
        indices = self._to_linear_index[latitude_index, longitude_index]
        if valid_only:
            indices = indices[indices>=0]
        return indices

    def direction_step(self):
        return wrapped_difference(
            numpy.diff(self.directions,append=self.directions[0]),period=360
        )

    def frequency_step(self):
        delta = numpy.diff(self.frequencies,
                       prepend=self.frequencies[0]/self._growth_factor,
                       append=self.frequencies[-1]*self._growth_factor
        )
        return (delta[0:-1] + delta[1:]) / 2

    def extract(self, s: slice, var: numpy.ndarray):
        """
        Extract linear indexed data (with indices indicated by the slice) from
        a 2D array (lat,lon)
        :param s: slice instance
        :param var: 2D ndarray
        :return: 1D ndarray of length slice.stop-slice.start
        """
        return var[self.latitude_index(s), self.longitude_index(s)]

    def project(self, lon_slice: slice, lat_slice: slice,
                var: numpy.ndarray, except_val=numpy.nan):
        """
        Project linear indexed data onto a 2d array (lat ,lon).

        :param lon_slice: slice of ilon indices we want
        :param lat_slice: slice of ilat indices we want
        :param var:
        :param except_val:
        :return:
        """
        ind = self._to_linear_index[lat_slice, lon_slice]
        mask = ind >= 0
        ind = ind[mask]
        ilon = self.longitude_index(ind)
        ilat = self.latitude_index(ind)

        nlon = len(self.longitude[lon_slice])
        nlat = len(self.latitude[lat_slice])

        # this will not work for stepsizes != 1
        out = numpy.zeros((nlat, nlon), dtype=var.dtype) + except_val
        out[ilat, ilon] = var[ind]
        return out

    def set_linear_data(self, lon_slice: slice, lat_slice: slice,
                        linear_data: numpy.ndarray, data: numpy.ndarray):
        """
        Update data stored in a linear array with data from a 2d array.
        :param lon_slice: longitude indices
        :param lat_slice: latitude indices
        :param linear_data: linear data array to update
        :param data: 2d Array
        :return:
        """

        ind = self._to_linear_index[lat_slice, lon_slice]
        ind = ind[ind >= 0]
        ilon = self.longitude_index(ind)
        ilat = self.latitude_index(ind)
        linear_data[ind] = data[ilat, ilon]
        return linear_data


class LinearIndexedGridData:
    """
    A simple class to interact with data that is stored in a packed linear
    fasion in wavewatch. Wavewatch uses a 2d grid (lat,lon) but stores spatial
    point in a 1d indexed array- where _only_ ocean/sea/lake points are
    included. To refer back to the original 2d tuple we therefore need a trans-
    lation array relating index -> ilat,ilon and ilat,ilon -> index. Note that
    for the second mapping not all points are defined (if ilat,ilon is a land
    point).

    Usage- to get data from the array we instantiate the object:

    linear_indexed_data = LinearIndexedGridData( ... )

    and subsequently use linear_indexed_data[linear_index] to get data from
    a linear index, or linear_indexed_data[ilat,ilon] to retrieve it from a
    specific lat lon. Assignment works similarly.
    """
    def __init__(self, linear_indexed_data: numpy.ndarray, grid: Grid):
        """
        :param linear_indexed_data: Data stored in packed linear fashion
        :param grid: the associated grid object
        """
        self._linear_indexed_data = linear_indexed_data
        self._grid = grid

    def __getitem__(self, *item) -> numpy.ndarray:
        if len(item) == 2:
            item = [_to_slice(x) for x in item]
            return self._grid.project(item[0], item[1],
                                      self._linear_indexed_data)
        elif isinstance(item[0], (Sequence, numpy.ndarray)):
            return self._linear_indexed_data[item[0]]

        else:
            return self._linear_indexed_data[_to_slice(item[0])]

    def __setitem__(self, *args):
        if len(args) == 3:
            item = [_to_slice(x) for x in args[:2]]
            self._grid.set_linear_data(item[0], item[1],
                                       self._linear_indexed_data, args[2])
        else:
            return self._linear_indexed_data[_to_slice(args[0])]


def read_model_definition(
        resource: Resource,
        byte_order: Literal["<", ">", "="] = '<'
) -> Tuple[Grid, LinearIndexedGridData, numpy.ndarray]:

    """
    Read a ww3 model definition file, see comments on top of this file for
    further info.

    :param resource: an instance of a resource.
    :param byte_order: which byte order data is stored. Sofar currently uses
        little endian (<).
    :return: the grid, depth and the landmask.
    """
    def jump(stream, number, start):
        if start is None:
            _ = stream.read(number)
        else:
            stream.seek(start + number)

    stream = BytesIO(resource.read())

    fort_char = FortranCharacter(endianness=byte_order)
    fort_int = FortranInt(endianness=byte_order)
    fort_float = FortranFloat(endianness=byte_order)

    data = {}
    start_of_record_absolute = 0

    # The need to read the start of record (_sor) and end of record (_eor) is
    # a quirk of the sequential unformatted IO that ww3 uses to write this
    # file and our use of gfortran to compile. Specifically:
    #
    # https://gcc.gnu.org/
    #    onlinedocs/gfortran/File-format-of-unformatted-sequential-files.html
    #
    # Each record in the file contains a leading
    # start of record (sor) 4 byte marker indicating size N of the record,
    # the data written (N bytes), and a trailing end of record (eor) marker again
    # indicating the size of the data. No information on the type of the data is
    # provided, and this has to be known in advance. E.g. mulitple different
    # information types can (and are) stored as part of the same record.
    _sor = fort_int.unpack(stream, 1)[0]  # Fortran record opening
    data['id'] = fort_char.unpack(stream, 35)  # Line 70
    data['grid_name'] = fort_char.unpack(stream, 10)  # Line 70  # noqa: F841
    data['nlon'] = fort_int.unpack(stream, 1)[0]  # Number of longitude points
    data['nlat'] = fort_int.unpack(stream, 1)[0]  # Number of longitude points
    data['num_sea'] = fort_int.unpack(stream, 1)[0]  # Number of sea points
    data['num_dir'] = fort_int.unpack(stream, 1)[
        0]  # Number of directional points
    data['num_freq'] = fort_int.unpack(stream, 1)[
        0]  # Number of wavenumber points

    # Number of input bound points (see w3odatd.ftn, line 219) and
    # Number of files for output bound data (see w3odatd.ftn, line 220)
    # (3 int total, not used here, skipping, jumping to eor
    _ = jump(stream, _sor + 4, 0)
    _eor = fort_int.unpack(stream, 1)[0]
    assert _sor == _eor

    _sor = fort_int.unpack(stream, 1)[0]
    _ = jump(stream, _sor, None)
    _eor = fort_int.unpack(stream, 1)[0]
    assert _sor == _eor

    # assert fort_int.unpack(stream,1)[0] == _sor

    # Enter the section for W3GDAT (line 582)
    data["grid_type"], data["flagll"], data["iclose"] = \
        fort_int.unpack(stream, None, unformatted_sequential=True)
    (data["longitude_stepsize"], data["latitude_stepsize"],
     data["longitude_start"], data["latitude_start"]) = \
        fort_float.unpack(stream, None, unformatted_sequential=True)

    _sor = fort_int.unpack(stream, 1)[0]
    start_of_record_absolute = stream.tell()
    dtype_float = numpy.dtype('float32').newbyteorder(byte_order)
    dtype_int = numpy.dtype('int32').newbyteorder(byte_order)

    # First the bottom grid. The bottom grid is only stored on the
    # computational grid
    data['bottom_datum'] = \
        numpy.frombuffer(
            stream.read(data['num_sea'] * 4),
            count=data['num_sea'], dtype=dtype_float)

    # Next the mask layer. The masked layer is stored everywhere. Note that it
    # is stored with latitude as the fast axis- here we retrieve as longitude
    # as the fast axis (transpose)
    num_points = data['nlon'] * data['nlat']
    data['mask'] = \
        numpy.reshape(
            numpy.frombuffer(stream.read(num_points * 4), dtype=dtype_int),
            (data['nlon'], data['nlat'])
        ).transpose()

    # Next the mapping that relates ix,iy -> ns. Subtract 1 to account for 0
    # based indexing (vs 1 in Fortran). Note that it is stored with latitude as
    # the fast axis- here we retrieve as longitude as the fast axis (transpose)
    data['to_linear_index'] = \
        numpy.reshape(
            numpy.frombuffer(stream.read(num_points * 4), dtype=dtype_int),
            (data['nlon'], data['nlat'])
        ).transpose() - 1

    num_points = data['num_sea']
    data['to_point_index'] = \
        numpy.reshape(
            numpy.frombuffer(stream.read(num_points * 8), dtype=dtype_int),
            (2, num_points)
        ) - 1

    # FROM: w3gdatmd.ftn line 160
    #      TRFLAG    Int.  Public   Flag for use of transparencies
    #                                0: No sub-grid obstacles.
    #                                1: Obstructions at cell boundaries.
    #                                2: Obstructions at cell centers.
    #                                3: Like 1 with continuous ice.
    #                                4: Like 2 with continuous ice.
    data['tr_flag'] = fort_int.unpack(stream, 1)[0]
    _ = jump(stream, _sor, start_of_record_absolute)
    _eor = fort_int.unpack(stream, 1)[0]
    assert _sor == _eor

    # we jump three values associated with TRFLAG
    for ii in range(0, 3):
        _ = fort_int.unpack(stream, None, True)

    # Spectral parameters
    # For descriptions; see w3gdatmd.ftn
    num_spec_points = data['num_freq'] * data['num_dir']

    # I do not fully understand why arrays we ignore are larger.. but we
    # ignore them
    byte_size_full = (num_spec_points + data['num_dir']) * 4

    # ignore: "mapwn, mapth"
    _sor = fort_int.unpack(stream, 1)[0]
    _ = stream.read(byte_size_full * 2)
    data['direction_step_size'] = fort_float.unpack(stream, 1)[0]
    data['direction_degree'] = numpy.frombuffer(
        stream.read(data['num_dir'] * 4), dtype=dtype_float
    ) * 180 / numpy.pi

    # ignore: "stuff"
    _ = stream.read(byte_size_full * 5)

    #
    data['freq_mult_fac'] = fort_float.unpack(stream, 1)[0]
    data['start_frequency'] = fort_float.unpack(stream, 1)[0]

    # Note they store an extra frequency before start, and at the end, I
    # assume this is for convinience in calculating delta's- either way,
    # ignored here.
    data['frequency_hertz'] = (numpy.frombuffer(
        stream.read(data['num_freq'] * 4 + 8), dtype=dtype_float
    ) / numpy.pi / 2)[1:-1]

    latitude = numpy.linspace(
        data['latitude_start'],
        data['latitude_start'] + data['latitude_stepsize']
        * data['nlat'], data['nlat'],
        endpoint=True
    )

    longitude = numpy.linspace(
        data['longitude_start'],
        data['longitude_start'] + data['longitude_stepsize']
        * data['nlon'], data['nlon'],
        endpoint=False
    )
    grid = Grid(number_of_spatial_points=data['num_sea'],
                frequencies=data['frequency_hertz'],
                directions=data['direction_degree'],
                latitude=latitude,
                longitude=longitude,
                _to_linear_index=data['to_linear_index'],
                _to_point_index=data['to_point_index'])

    depth = LinearIndexedGridData(- data['bottom_datum'], grid)

    return grid, depth, data['mask']


def _to_slice(val: Union[slice, int]) -> slice:
    if not isinstance(val, (slice, int)):
        print(type(val), val)
        raise ValueError('Only slice or int supported as index.')

    if isinstance(val, int):
        return slice(val, val + 1, 1)
    else:
        return val
